# Copyright (c) 2016 Clinton Knight
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt

from manila import exception
from manila.share.drivers.netapp.dataontap.client import api as netapp_api
from manila.share.drivers.netapp.dataontap.cluster_mode import performance
from manila import test
from manila.tests.share.drivers.netapp.dataontap import fakes as fake


@ddt.ddt
class PerformanceLibraryTestCase(test.TestCase):

    def setUp(self):
        super(PerformanceLibraryTestCase, self).setUp()

        with mock.patch.object(performance.PerformanceLibrary,
                               '_init_counter_info'):
            self.zapi_client = mock.Mock()
            self.perf_library = performance.PerformanceLibrary(
                self.zapi_client)
            self.perf_library.system_object_name = 'system'
            self.perf_library.avg_processor_busy_base_counter_name = (
                'cpu_elapsed_time1')

        self._set_up_fake_pools()

    def _set_up_fake_pools(self):

        self.fake_volumes = {
            'pool1': {
                'netapp_aggregate': 'aggr1',
            },
            'pool2': {
                'netapp_aggregate': 'aggr2',
            },
            'pool3': {
                'netapp_aggregate': 'aggr2',
            },
        }
        self.fake_aggregates = {
            'pool4': {
                'netapp_aggregate': 'aggr3',
            },
            'flexgroup_pool': {
                'netapp_aggregate': 'aggr1 aggr2',
                'netapp_flexgroup': True,
            },
        }

        self.fake_aggr_names = ['aggr1', 'aggr2', 'aggr3']
        self.fake_nodes = ['node1', 'node2']
        self.fake_aggr_node_map = {
            'aggr1': 'node1',
            'aggr2': 'node2',
            'aggr3': 'node2',
        }

    def _get_fake_counters(self):

        return {
            'node1': list(range(11, 21)),
            'node2': list(range(21, 31)),
        }

    def test_init(self):

        mock_zapi_client = mock.Mock()
        mock_init_counter_info = self.mock_object(
            performance.PerformanceLibrary, '_init_counter_info')

        library = performance.PerformanceLibrary(mock_zapi_client)

        self.assertEqual(mock_zapi_client, library.zapi_client)
        mock_init_counter_info.assert_called_once_with()

    def test_init_counter_info_not_supported(self):

        self.zapi_client.features.SYSTEM_METRICS = False
        self.zapi_client.features.SYSTEM_CONSTITUENT_METRICS = False
        mock_get_base_counter_name = self.mock_object(
            self.perf_library, '_get_base_counter_name')

        self.perf_library._init_counter_info()

        self.assertIsNone(self.perf_library.system_object_name)
        self.assertIsNone(
            self.perf_library.avg_processor_busy_base_counter_name)
        self.assertFalse(mock_get_base_counter_name.called)

    @ddt.data({
        'system_constituent': False,
        'base_counter': 'cpu_elapsed_time1',
    }, {
        'system_constituent': True,
        'base_counter': 'cpu_elapsed_time',
    })
    @ddt.unpack
    def test_init_counter_info_api_error(self, system_constituent,
                                         base_counter):

        self.zapi_client.features.SYSTEM_METRICS = True
        self.zapi_client.features.SYSTEM_CONSTITUENT_METRICS = (
            system_constituent)
        self.mock_object(self.perf_library,
                         '_get_base_counter_name',
                         mock.Mock(side_effect=netapp_api.NaApiError))

        self.perf_library._init_counter_info()

        self.assertEqual(
            base_counter,
            self.perf_library.avg_processor_busy_base_counter_name)

    def test_init_counter_info_system(self):

        self.zapi_client.features.SYSTEM_METRICS = True
        self.zapi_client.features.SYSTEM_CONSTITUENT_METRICS = False
        mock_get_base_counter_name = self.mock_object(
            self.perf_library, '_get_base_counter_name',
            mock.Mock(return_value='cpu_elapsed_time1'))

        self.perf_library._init_counter_info()

        self.assertEqual('system', self.perf_library.system_object_name)
        self.assertEqual(
            'cpu_elapsed_time1',
            self.perf_library.avg_processor_busy_base_counter_name)
        mock_get_base_counter_name.assert_called_once_with(
            'system', 'avg_processor_busy')

    def test_init_counter_info_system_constituent(self):

        self.zapi_client.features.SYSTEM_METRICS = False
        self.zapi_client.features.SYSTEM_CONSTITUENT_METRICS = True
        mock_get_base_counter_name = self.mock_object(
            self.perf_library, '_get_base_counter_name',
            mock.Mock(return_value='cpu_elapsed_time'))

        self.perf_library._init_counter_info()

        self.assertEqual('system:constituent',
                         self.perf_library.system_object_name)
        self.assertEqual(
            'cpu_elapsed_time',
            self.perf_library.avg_processor_busy_base_counter_name)
        mock_get_base_counter_name.assert_called_once_with(
            'system:constituent', 'avg_processor_busy')

    def test_update_performance_cache(self):

        self.perf_library.performance_counters = self._get_fake_counters()
        mock_get_aggregates_for_pools = self.mock_object(
            self.perf_library, '_get_aggregates_for_pools',
            mock.Mock(return_value=self.fake_aggr_names))
        mock_get_nodes_for_aggregates = self.mock_object(
            self.perf_library, '_get_nodes_for_aggregates',
            mock.Mock(return_value=(self.fake_nodes,
                                    self.fake_aggr_node_map)))
        mock_get_node_utilization_counters = self.mock_object(
            self.perf_library, '_get_node_utilization_counters',
            mock.Mock(side_effect=[21, 31]))
        mock_get_node_utilization = self.mock_object(
            self.perf_library, '_get_node_utilization',
            mock.Mock(side_effect=[25, 75]))

        self.perf_library.update_performance_cache(self.fake_volumes,
                                                   self.fake_aggregates)

        expected_performance_counters = {
            'node1': list(range(12, 22)),
            'node2': list(range(22, 32)),
        }
        self.assertEqual(expected_performance_counters,
                         self.perf_library.performance_counters)

        expected_pool_utilization = {
            'pool1': 25,
            'pool2': 75,
            'pool3': 75,
            'pool4': 75,
            'flexgroup_pool': performance.DEFAULT_UTILIZATION,
        }
        self.assertEqual(expected_pool_utilization,
                         self.perf_library.pool_utilization)

        mock_get_aggregates_for_pools.assert_called_once_with(
            self.fake_volumes, self.fake_aggregates)
        mock_get_nodes_for_aggregates.assert_called_once_with(
            self.fake_aggr_names)
        mock_get_node_utilization_counters.assert_has_calls([
            mock.call('node1'), mock.call('node2')])
        mock_get_node_utilization.assert_has_calls([
            mock.call(12, 21, 'node1'), mock.call(22, 31, 'node2')])

    def test_update_performance_cache_first_pass(self):

        mock_get_aggregates_for_pools = self.mock_object(
            self.perf_library, '_get_aggregates_for_pools',
            mock.Mock(return_value=self.fake_aggr_names))
        mock_get_nodes_for_aggregates = self.mock_object(
            self.perf_library, '_get_nodes_for_aggregates',
            mock.Mock(return_value=(self.fake_nodes,
                                    self.fake_aggr_node_map)))
        mock_get_node_utilization_counters = self.mock_object(
            self.perf_library, '_get_node_utilization_counters',
            mock.Mock(side_effect=[11, 21]))
        mock_get_node_utilization = self.mock_object(
            self.perf_library, '_get_node_utilization',
            mock.Mock(side_effect=[25, 75]))

        self.perf_library.update_performance_cache(self.fake_volumes,
                                                   self.fake_aggregates)

        expected_performance_counters = {'node1': [11], 'node2': [21]}
        self.assertEqual(expected_performance_counters,
                         self.perf_library.performance_counters)

        expected_pool_utilization = {
            'pool1': performance.DEFAULT_UTILIZATION,
            'pool2': performance.DEFAULT_UTILIZATION,
            'pool3': performance.DEFAULT_UTILIZATION,
            'pool4': performance.DEFAULT_UTILIZATION,
            'flexgroup_pool': performance.DEFAULT_UTILIZATION,
        }
        self.assertEqual(expected_pool_utilization,
                         self.perf_library.pool_utilization)

        mock_get_aggregates_for_pools.assert_called_once_with(
            self.fake_volumes, self.fake_aggregates)
        mock_get_nodes_for_aggregates.assert_called_once_with(
            self.fake_aggr_names)
        mock_get_node_utilization_counters.assert_has_calls([
            mock.call('node1'), mock.call('node2')])
        self.assertFalse(mock_get_node_utilization.called)

    def test_update_performance_cache_unknown_nodes(self):

        self.perf_library.performance_counters = self._get_fake_counters()
        mock_get_aggregates_for_pools = self.mock_object(
            self.perf_library, '_get_aggregates_for_pools',
            mock.Mock(return_value=self.fake_aggr_names))
        mock_get_nodes_for_aggregates = self.mock_object(
            self.perf_library, '_get_nodes_for_aggregates',
            mock.Mock(return_value=([], {})))
        mock_get_node_utilization_counters = self.mock_object(
            self.perf_library, '_get_node_utilization_counters',
            mock.Mock(side_effect=[11, 21]))
        mock_get_node_utilization = self.mock_object(
            self.perf_library, '_get_node_utilization',
            mock.Mock(side_effect=[25, 75]))

        self.perf_library.update_performance_cache(self.fake_volumes,
                                                   self.fake_aggregates)

        self.assertEqual(self._get_fake_counters(),
                         self.perf_library.performance_counters)

        expected_pool_utilization = {
            'pool1': performance.DEFAULT_UTILIZATION,
            'pool2': performance.DEFAULT_UTILIZATION,
            'pool3': performance.DEFAULT_UTILIZATION,
            'pool4': performance.DEFAULT_UTILIZATION,
            'flexgroup_pool': performance.DEFAULT_UTILIZATION,
        }
        self.assertEqual(expected_pool_utilization,
                         self.perf_library.pool_utilization)

        mock_get_aggregates_for_pools.assert_called_once_with(
            self.fake_volumes, self.fake_aggregates)
        mock_get_nodes_for_aggregates.assert_called_once_with(
            self.fake_aggr_names)
        self.assertFalse(mock_get_node_utilization_counters.called)
        self.assertFalse(mock_get_node_utilization.called)

    def test_update_performance_cache_counters_unavailable(self):

        self.perf_library.performance_counters = self._get_fake_counters()
        mock_get_aggregates_for_pools = self.mock_object(
            self.perf_library, '_get_aggregates_for_pools',
            mock.Mock(return_value=self.fake_aggr_names))
        mock_get_nodes_for_aggregates = self.mock_object(
            self.perf_library, '_get_nodes_for_aggregates',
            mock.Mock(return_value=(self.fake_nodes,
                                    self.fake_aggr_node_map)))
        mock_get_node_utilization_counters = self.mock_object(
            self.perf_library, '_get_node_utilization_counters',
            mock.Mock(side_effect=[None, None]))
        mock_get_node_utilization = self.mock_object(
            self.perf_library, '_get_node_utilization',
            mock.Mock(side_effect=[25, 75]))

        self.perf_library.update_performance_cache(self.fake_volumes,
                                                   self.fake_aggregates)

        self.assertEqual(self._get_fake_counters(),
                         self.perf_library.performance_counters)

        expected_pool_utilization = {
            'pool1': performance.DEFAULT_UTILIZATION,
            'pool2': performance.DEFAULT_UTILIZATION,
            'pool3': performance.DEFAULT_UTILIZATION,
            'pool4': performance.DEFAULT_UTILIZATION,
            'flexgroup_pool': performance.DEFAULT_UTILIZATION,
        }
        self.assertEqual(expected_pool_utilization,
                         self.perf_library.pool_utilization)

        mock_get_aggregates_for_pools.assert_called_once_with(
            self.fake_volumes, self.fake_aggregates)
        mock_get_nodes_for_aggregates.assert_called_once_with(
            self.fake_aggr_names)
        mock_get_node_utilization_counters.assert_has_calls([
            mock.call('node1'), mock.call('node2')])
        self.assertFalse(mock_get_node_utilization.called)

    def test_update_performance_cache_not_supported(self):

        self.zapi_client.features.SYSTEM_METRICS = False
        self.zapi_client.features.SYSTEM_CONSTITUENT_METRICS = False

        mock_get_aggregates_for_pools = self.mock_object(
            self.perf_library, '_get_aggregates_for_pools')

        self.perf_library.update_performance_cache(self.fake_volumes,
                                                   self.fake_aggregates)

        expected_performance_counters = {}
        self.assertEqual(expected_performance_counters,
                         self.perf_library.performance_counters)

        expected_pool_utilization = {}
        self.assertEqual(expected_pool_utilization,
                         self.perf_library.pool_utilization)

        self.assertFalse(mock_get_aggregates_for_pools.called)

    @ddt.data({'pool': 'pool1', 'expected': 10.0},
              {'pool': 'pool3', 'expected': performance.DEFAULT_UTILIZATION})
    @ddt.unpack
    def test_get_node_utilization_for_pool(self, pool, expected):

        self.perf_library.pool_utilization = {'pool1': 10.0, 'pool2': 15.0}

        result = self.perf_library.get_node_utilization_for_pool(pool)

        self.assertAlmostEqual(expected, result)

    def test__update_for_failover(self):
        self.mock_object(self.perf_library, 'update_performance_cache')
        mock_client = mock.Mock(name='FAKE_ZAPI_CLIENT')

        self.perf_library.update_for_failover(mock_client,
                                              self.fake_volumes,
                                              self.fake_aggregates)

        self.assertEqual(mock_client, self.perf_library.zapi_client)
        self.perf_library.update_performance_cache.assert_called_once_with(
            self.fake_volumes, self.fake_aggregates)

    def test_get_aggregates_for_pools(self):

        result = self.perf_library._get_aggregates_for_pools(
            self.fake_volumes, self.fake_aggregates)

        expected_aggregate_names = ['aggr1', 'aggr2', 'aggr3']
        self.assertEqual(sorted(expected_aggregate_names), sorted(result))

    def test_get_nodes_for_aggregates(self):

        aggregate_names = ['aggr1', 'aggr2', 'aggr3']
        aggregate_nodes = ['node1', 'node2', 'node2']

        mock_get_node_for_aggregate = self.mock_object(
            self.zapi_client, 'get_node_for_aggregate',
            mock.Mock(side_effect=aggregate_nodes))

        result = self.perf_library._get_nodes_for_aggregates(aggregate_names)

        self.assertEqual(2, len(result))
        result_node_names, result_aggr_node_map = result

        expected_node_names = ['node1', 'node2']
        expected_aggr_node_map = dict(zip(aggregate_names, aggregate_nodes))
        self.assertEqual(sorted(expected_node_names),
                         sorted(result_node_names))
        self.assertEqual(expected_aggr_node_map, result_aggr_node_map)
        mock_get_node_for_aggregate.assert_has_calls([
            mock.call('aggr1'), mock.call('aggr2'), mock.call('aggr3')])

    def test_get_node_utilization_kahuna_overutilized(self):

        mock_get_kahuna_utilization = self.mock_object(
            self.perf_library, '_get_kahuna_utilization',
            mock.Mock(return_value=61.0))
        mock_get_average_cpu_utilization = self.mock_object(
            self.perf_library, '_get_average_cpu_utilization',
            mock.Mock(return_value=25.0))

        result = self.perf_library._get_node_utilization('fake1',
                                                         'fake2',
                                                         'fake_node')

        self.assertAlmostEqual(100.0, result)
        mock_get_kahuna_utilization.assert_called_once_with('fake1', 'fake2')
        self.assertFalse(mock_get_average_cpu_utilization.called)

    @ddt.data({'cpu': -0.01, 'cp_time': 10000, 'poll_time': 0},
              {'cpu': 1.01, 'cp_time': 0, 'poll_time': 1000},
              {'cpu': 0.50, 'cp_time': 0, 'poll_time': 0})
    @ddt.unpack
    def test_get_node_utilization_zero_time(self, cpu, cp_time, poll_time):

        mock_get_kahuna_utilization = self.mock_object(
            self.perf_library, '_get_kahuna_utilization',
            mock.Mock(return_value=59.0))
        mock_get_average_cpu_utilization = self.mock_object(
            self.perf_library, '_get_average_cpu_utilization',
            mock.Mock(return_value=cpu))
        mock_get_total_consistency_point_time = self.mock_object(
            self.perf_library, '_get_total_consistency_point_time',
            mock.Mock(return_value=cp_time))
        mock_get_consistency_point_p2_flush_time = self.mock_object(
            self.perf_library, '_get_consistency_point_p2_flush_time',
            mock.Mock(return_value=cp_time))
        mock_get_total_time = self.mock_object(
            self.perf_library, '_get_total_time',
            mock.Mock(return_value=poll_time))
        mock_get_adjusted_consistency_point_time = self.mock_object(
            self.perf_library, '_get_adjusted_consistency_point_time')

        result = self.perf_library._get_node_utilization('fake1',
                                                         'fake2',
                                                         'fake_node')

        expected = max(min(100.0, 100.0 * cpu), 0)
        self.assertEqual(expected, result)

        mock_get_kahuna_utilization.assert_called_once_with('fake1', 'fake2')
        mock_get_average_cpu_utilization.assert_called_once_with('fake1',
                                                                 'fake2')
        mock_get_total_consistency_point_time.assert_called_once_with('fake1',
                                                                      'fake2')
        mock_get_consistency_point_p2_flush_time.assert_called_once_with(
            'fake1', 'fake2')
        mock_get_total_time.assert_called_once_with('fake1',
                                                    'fake2',
                                                    'total_cp_msecs')
        self.assertFalse(mock_get_adjusted_consistency_point_time.called)

    @ddt.data({'cpu': 0.75, 'adjusted_cp_time': 8000, 'expected': 80},
              {'cpu': 0.80, 'adjusted_cp_time': 7500, 'expected': 80},
              {'cpu': 0.50, 'adjusted_cp_time': 11000, 'expected': 100})
    @ddt.unpack
    def test_get_node_utilization(self, cpu, adjusted_cp_time, expected):

        mock_get_kahuna_utilization = self.mock_object(
            self.perf_library, '_get_kahuna_utilization',
            mock.Mock(return_value=59.0))
        mock_get_average_cpu_utilization = self.mock_object(
            self.perf_library, '_get_average_cpu_utilization',
            mock.Mock(return_value=cpu))
        mock_get_total_consistency_point_time = self.mock_object(
            self.perf_library, '_get_total_consistency_point_time',
            mock.Mock(return_value=90.0))
        mock_get_consistency_point_p2_flush_time = self.mock_object(
            self.perf_library, '_get_consistency_point_p2_flush_time',
            mock.Mock(return_value=50.0))
        mock_get_total_time = self.mock_object(
            self.perf_library, '_get_total_time',
            mock.Mock(return_value=10000))
        mock_get_adjusted_consistency_point_time = self.mock_object(
            self.perf_library, '_get_adjusted_consistency_point_time',
            mock.Mock(return_value=adjusted_cp_time))

        result = self.perf_library._get_node_utilization('fake1',
                                                         'fake2',
                                                         'fake_node')

        self.assertEqual(expected, result)

        mock_get_kahuna_utilization.assert_called_once_with('fake1', 'fake2')
        mock_get_average_cpu_utilization.assert_called_once_with('fake1',
                                                                 'fake2')
        mock_get_total_consistency_point_time.assert_called_once_with('fake1',
                                                                      'fake2')
        mock_get_consistency_point_p2_flush_time.assert_called_once_with(
            'fake1', 'fake2')
        mock_get_total_time.assert_called_once_with('fake1',
                                                    'fake2',
                                                    'total_cp_msecs')
        mock_get_adjusted_consistency_point_time.assert_called_once_with(
            90.0, 50.0)

    def test_get_node_utilization_calculation_error(self):

        self.mock_object(self.perf_library,
                         '_get_kahuna_utilization',
                         mock.Mock(return_value=59.0))
        self.mock_object(self.perf_library,
                         '_get_average_cpu_utilization',
                         mock.Mock(return_value=25.0))
        self.mock_object(self.perf_library,
                         '_get_total_consistency_point_time',
                         mock.Mock(return_value=90.0))
        self.mock_object(self.perf_library,
                         '_get_consistency_point_p2_flush_time',
                         mock.Mock(return_value=50.0))
        self.mock_object(self.perf_library,
                         '_get_total_time',
                         mock.Mock(return_value=10000))
        self.mock_object(self.perf_library,
                         '_get_adjusted_consistency_point_time',
                         mock.Mock(side_effect=ZeroDivisionError))

        result = self.perf_library._get_node_utilization('fake1',
                                                         'fake2',
                                                         'fake_node')

        self.assertEqual(performance.DEFAULT_UTILIZATION, result)
        (self.perf_library._get_adjusted_consistency_point_time.
            assert_called_once_with(mock.ANY, mock.ANY))

    def test_get_kahuna_utilization(self):

        mock_get_performance_counter = self.mock_object(
            self.perf_library,
            '_get_performance_counter_average_multi_instance',
            mock.Mock(return_value=[0.2, 0.3]))

        result = self.perf_library._get_kahuna_utilization('fake_t1',
                                                           'fake_t2')

        self.assertAlmostEqual(50.0, result)
        mock_get_performance_counter.assert_called_once_with(
            'fake_t1', 'fake_t2', 'domain_busy:kahuna',
            'processor_elapsed_time')

    def test_get_average_cpu_utilization(self):

        mock_get_performance_counter_average = self.mock_object(
            self.perf_library, '_get_performance_counter_average',
            mock.Mock(return_value=0.45))

        result = self.perf_library._get_average_cpu_utilization('fake_t1',
                                                                'fake_t2')

        self.assertAlmostEqual(0.45, result)
        mock_get_performance_counter_average.assert_called_once_with(
            'fake_t1', 'fake_t2', 'avg_processor_busy', 'cpu_elapsed_time1')

    def test_get_total_consistency_point_time(self):

        mock_get_performance_counter_delta = self.mock_object(
            self.perf_library, '_get_performance_counter_delta',
            mock.Mock(return_value=500))

        result = self.perf_library._get_total_consistency_point_time(
            'fake_t1', 'fake_t2')

        self.assertEqual(500, result)
        mock_get_performance_counter_delta.assert_called_once_with(
            'fake_t1', 'fake_t2', 'total_cp_msecs')

    def test_get_consistency_point_p2_flush_time(self):

        mock_get_performance_counter_delta = self.mock_object(
            self.perf_library, '_get_performance_counter_delta',
            mock.Mock(return_value=500))

        result = self.perf_library._get_consistency_point_p2_flush_time(
            'fake_t1', 'fake_t2')

        self.assertEqual(500, result)
        mock_get_performance_counter_delta.assert_called_once_with(
            'fake_t1', 'fake_t2', 'cp_phase_times:p2_flush')

    def test_get_total_time(self):

        mock_find_performance_counter_timestamp = self.mock_object(
            self.perf_library, '_find_performance_counter_timestamp',
            mock.Mock(side_effect=[100, 105]))

        result = self.perf_library._get_total_time('fake_t1',
                                                   'fake_t2',
                                                   'fake_counter')

        self.assertEqual(5000, result)
        mock_find_performance_counter_timestamp.assert_has_calls([
            mock.call('fake_t1', 'fake_counter'),
            mock.call('fake_t2', 'fake_counter')])

    def test_get_adjusted_consistency_point_time(self):

        result = self.perf_library._get_adjusted_consistency_point_time(
            500, 200)

        self.assertAlmostEqual(360.0, result)

    def test_get_performance_counter_delta(self):

        result = self.perf_library._get_performance_counter_delta(
            fake.COUNTERS_T1, fake.COUNTERS_T2, 'total_cp_msecs')

        self.assertEqual(1482, result)

    def test_get_performance_counter_average(self):

        result = self.perf_library._get_performance_counter_average(
            fake.COUNTERS_T1, fake.COUNTERS_T2, 'domain_busy:kahuna',
            'processor_elapsed_time', 'processor0')

        self.assertAlmostEqual(0.00281954360981, result)

    def test_get_performance_counter_average_multi_instance(self):

        result = (
            self.perf_library._get_performance_counter_average_multi_instance(
                fake.COUNTERS_T1, fake.COUNTERS_T2, 'domain_busy:kahuna',
                'processor_elapsed_time'))

        expected = [0.002819543609809441, 0.0033421611147606135]
        self.assertAlmostEqual(expected, result)

    def test_find_performance_counter_value(self):

        result = self.perf_library._find_performance_counter_value(
            fake.COUNTERS_T1, 'domain_busy:kahuna',
            instance_name='processor0')

        self.assertEqual('2712467226', result)

    def test_find_performance_counter_value_not_found(self):

        self.assertRaises(
            exception.NotFound,
            self.perf_library._find_performance_counter_value,
            fake.COUNTERS_T1, 'invalid', instance_name='processor0')

    def test_find_performance_counter_timestamp(self):

        result = self.perf_library._find_performance_counter_timestamp(
            fake.COUNTERS_T1, 'domain_busy')

        self.assertEqual('1453573777', result)

    def test_find_performance_counter_timestamp_not_found(self):

        self.assertRaises(
            exception.NotFound,
            self.perf_library._find_performance_counter_timestamp,
            fake.COUNTERS_T1, 'invalid', instance_name='processor0')

    def test_expand_performance_array(self):

        counter_info = {
            'labels': ['idle', 'kahuna', 'storage', 'exempt'],
            'name': 'domain_busy',
        }
        self.zapi_client.get_performance_counter_info = mock.Mock(
            return_value=counter_info)

        counter = {
            'node-name': 'cluster1-01',
            'instance-uuid': 'cluster1-01:kernel:processor0',
            'domain_busy': '969142314286,2567571412,2131582146,5383861579',
            'instance-name': 'processor0',
            'timestamp': '1453512244',
        }
        self.perf_library._expand_performance_array('wafl',
                                                    'domain_busy',
                                                    counter)

        modified_counter = {
            'node-name': 'cluster1-01',
            'instance-uuid': 'cluster1-01:kernel:processor0',
            'domain_busy': '969142314286,2567571412,2131582146,5383861579',
            'instance-name': 'processor0',
            'timestamp': '1453512244',
            'domain_busy:idle': '969142314286',
            'domain_busy:kahuna': '2567571412',
            'domain_busy:storage': '2131582146',
            'domain_busy:exempt': '5383861579',
        }
        self.assertEqual(modified_counter, counter)

    def test_get_base_counter_name(self):

        counter_info = {
            'base-counter': 'cpu_elapsed_time',
            'labels': [],
            'name': 'avg_processor_busy',
        }
        self.zapi_client.get_performance_counter_info = mock.Mock(
            return_value=counter_info)

        result = self.perf_library._get_base_counter_name(
            'system:constituent', 'avg_processor_busy')

        self.assertEqual('cpu_elapsed_time', result)

    def test_get_node_utilization_counters(self):

        mock_get_node_utilization_system_counters = self.mock_object(
            self.perf_library, '_get_node_utilization_system_counters',
            mock.Mock(return_value=['A', 'B', 'C']))
        mock_get_node_utilization_wafl_counters = self.mock_object(
            self.perf_library, '_get_node_utilization_wafl_counters',
            mock.Mock(return_value=['D', 'E', 'F']))
        mock_get_node_utilization_processor_counters = self.mock_object(
            self.perf_library, '_get_node_utilization_processor_counters',
            mock.Mock(return_value=['G', 'H', 'I']))

        result = self.perf_library._get_node_utilization_counters(fake.NODE)

        expected = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I']
        self.assertEqual(expected, result)

        mock_get_node_utilization_system_counters.assert_called_once_with(
            fake.NODE)
        mock_get_node_utilization_wafl_counters.assert_called_once_with(
            fake.NODE)
        mock_get_node_utilization_processor_counters.assert_called_once_with(
            fake.NODE)

    def test_get_node_utilization_counters_api_error(self):

        self.mock_object(self.perf_library,
                         '_get_node_utilization_system_counters',
                         mock.Mock(side_effect=netapp_api.NaApiError))

        result = self.perf_library._get_node_utilization_counters(fake.NODE)

        self.assertIsNone(result)

    def test_get_node_utilization_system_counters(self):

        mock_get_performance_instance_uuids = self.mock_object(
            self.zapi_client, 'get_performance_instance_uuids',
            mock.Mock(return_value=fake.SYSTEM_INSTANCE_UUIDS))
        mock_get_performance_counters = self.mock_object(
            self.zapi_client, 'get_performance_counters',
            mock.Mock(return_value=fake.SYSTEM_COUNTERS))

        result = self.perf_library._get_node_utilization_system_counters(
            fake.NODE)

        self.assertEqual(fake.SYSTEM_COUNTERS, result)

        mock_get_performance_instance_uuids.assert_called_once_with(
            'system', fake.NODE)
        mock_get_performance_counters.assert_called_once_with(
            'system', fake.SYSTEM_INSTANCE_UUIDS,
            ['avg_processor_busy', 'cpu_elapsed_time1', 'cpu_elapsed_time'])

    def test_get_node_utilization_wafl_counters(self):

        mock_get_performance_instance_uuids = self.mock_object(
            self.zapi_client, 'get_performance_instance_uuids',
            mock.Mock(return_value=fake.WAFL_INSTANCE_UUIDS))
        mock_get_performance_counters = self.mock_object(
            self.zapi_client, 'get_performance_counters',
            mock.Mock(return_value=fake.WAFL_COUNTERS))
        mock_get_performance_counter_info = self.mock_object(
            self.zapi_client, 'get_performance_counter_info',
            mock.Mock(return_value=fake.WAFL_CP_PHASE_TIMES_COUNTER_INFO))

        result = self.perf_library._get_node_utilization_wafl_counters(
            fake.NODE)

        self.assertEqual(fake.EXPANDED_WAFL_COUNTERS, result)

        mock_get_performance_instance_uuids.assert_called_once_with(
            'wafl', fake.NODE)
        mock_get_performance_counters.assert_called_once_with(
            'wafl', fake.WAFL_INSTANCE_UUIDS,
            ['total_cp_msecs', 'cp_phase_times'])
        mock_get_performance_counter_info.assert_called_once_with(
            'wafl', 'cp_phase_times')

    def test_get_node_utilization_processor_counters(self):

        mock_get_performance_instance_uuids = self.mock_object(
            self.zapi_client, 'get_performance_instance_uuids',
            mock.Mock(return_value=fake.PROCESSOR_INSTANCE_UUIDS))
        mock_get_performance_counters = self.mock_object(
            self.zapi_client, 'get_performance_counters',
            mock.Mock(return_value=fake.PROCESSOR_COUNTERS))
        self.mock_object(
            self.zapi_client, 'get_performance_counter_info',
            mock.Mock(return_value=fake.PROCESSOR_DOMAIN_BUSY_COUNTER_INFO))

        result = self.perf_library._get_node_utilization_processor_counters(
            fake.NODE)

        self.assertEqual(fake.EXPANDED_PROCESSOR_COUNTERS, result)

        mock_get_performance_instance_uuids.assert_called_once_with(
            'processor', fake.NODE)
        mock_get_performance_counters.assert_called_once_with(
            'processor', fake.PROCESSOR_INSTANCE_UUIDS,
            ['domain_busy', 'processor_elapsed_time'])



# Copyright (c) 2016 Hitachi Data Systems.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add_share_snapshot_access

Revision ID: a77e2ad5012d
Revises: e1949a93157a
Create Date: 2016-07-15 13:32:19.417771

"""

# revision identifiers, used by Alembic.
revision = 'a77e2ad5012d'
down_revision = 'e1949a93157a'

from manila.common import constants
from manila.db.migrations import utils

from alembic import op

import sqlalchemy as sa


def upgrade():
    op.create_table(
        'share_snapshot_access_map',
        sa.Column('id', sa.String(36), primary_key=True),
        sa.Column('created_at', sa.DateTime),
        sa.Column('updated_at', sa.DateTime),
        sa.Column('deleted_at', sa.DateTime),
        sa.Column('deleted', sa.String(36), default='False'),
        sa.Column('share_snapshot_id', sa.String(36),
                  sa.ForeignKey('share_snapshots.id',
                                name='ssam_snapshot_fk')),
        sa.Column('access_type', sa.String(255)),
        sa.Column('access_to', sa.String(255))
    )

    op.create_table(
        'share_snapshot_instance_access_map',
        sa.Column('id', sa.String(36), primary_key=True),
        sa.Column('created_at', sa.DateTime),
        sa.Column('updated_at', sa.DateTime),
        sa.Column('deleted_at', sa.DateTime),
        sa.Column('deleted', sa.String(36), default='False'),
        sa.Column('share_snapshot_instance_id', sa.String(36),
                  sa.ForeignKey('share_snapshot_instances.id',
                                name='ssiam_snapshot_instance_fk')),
        sa.Column('access_id', sa.String(36),
                  sa.ForeignKey('share_snapshot_access_map.id',
                                name='ssam_access_fk')),
        sa.Column('state', sa.String(255),
                  default=constants.ACCESS_STATE_QUEUED_TO_APPLY)
    )

    op.create_table(
        'share_snapshot_instance_export_locations',
        sa.Column('id', sa.String(36), primary_key=True),
        sa.Column('created_at', sa.DateTime),
        sa.Column('updated_at', sa.DateTime),
        sa.Column('deleted_at', sa.DateTime),
        sa.Column('deleted', sa.String(36), default='False'),
        sa.Column('share_snapshot_instance_id', sa.String(36),
                  sa.ForeignKey('share_snapshot_instances.id',
                                name='ssiel_snapshot_instance_fk')),
        sa.Column('path', sa.String(2000)),
        sa.Column('is_admin_only', sa.Boolean, default=False, nullable=False)
    )

    op.add_column('shares',
                  sa.Column('mount_snapshot_support', sa.Boolean,
                            default=False))

    connection = op.get_bind()
    shares_table = utils.load_table('shares', connection)

    # pylint: disable=no-value-for-parameter
    op.execute(
        shares_table.update().where(
            shares_table.c.deleted == 'False').values({
                'mount_snapshot_support': False,
            })
    )


def downgrade():
    op.drop_table('share_snapshot_instance_export_locations')
    op.drop_table('share_snapshot_instance_access_map')
    op.drop_table('share_snapshot_access_map')
    op.drop_column('shares', 'mount_snapshot_support')



# Copyright 2013 OpenStack Foundation
# Copyright 2015 Intel, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import datetime

from manila.api.openstack import api_version_request as api_version
from manila.common import constants
from manila.db.sqlalchemy import models
from manila.tests.db import fakes as db_fakes
from oslo_utils import uuidutils


def fake_share(**kwargs):

    share = {
        'id': 'fakeid',
        'name': 'fakename',
        'size': 1,
        'share_proto': 'fake_proto',
        'share_network_id': 'fake share network id',
        'share_server_id': 'fake share server id',
        'export_location': 'fake_location:/fake_share',
        'project_id': 'fake_project_uuid',
        'availability_zone': 'fake_az',
        'snapshot_support': 'True',
        'replication_type': None,
        'is_busy': False,
        'share_group_id': None,
        'instance': {
            'id': 'fake_share_instance_id',
            'host': 'fakehost',
            'share_type_id': '1',
            'share_network_id': 'fake share network id',
        },
        'mount_snapshot_support': False,
    }
    share.update(kwargs)
    return db_fakes.FakeModel(share)


def fake_share_instance(base_share=None, **kwargs):
    if base_share is None:
        share = fake_share()
    else:
        share = base_share

    share_instance = {
        'share_id': share['id'],
        'id': "fakeinstanceid",
        'status': "active",
        'host': 'fakehost',
        'share_network_id': 'fakesharenetworkid',
        'share_server_id': 'fakeshareserverid',
        'share_type_id': '1',
    }

    for attr in models.ShareInstance._proxified_properties:
        share_instance[attr] = getattr(share, attr, None)

    share_instance.update(kwargs)

    return db_fakes.FakeModel(share_instance)


def fake_share_type(**kwargs):

    share_type = {
        'id': "fakesharetype",
        'name': "fakesharetypename",
        'is_public': False,
        'extra_specs': {
            'driver_handles_share_servers': 'False',
        }
    }

    extra_specs = kwargs.pop('extra_specs', {})

    for key, value in extra_specs.items():
        share_type['extra_specs'][key] = value

    share_type.update(kwargs)

    return db_fakes.FakeModel(share_type)


def fake_snapshot(create_instance=False, **kwargs):

    instance_keys = ('instance_id', 'snapshot_id', 'share_instance_id',
                     'status', 'progress', 'provider_location')
    snapshot_keys = ('id', 'share_name', 'share_id', 'name', 'share_size',
                     'share_proto', 'instance', 'aggregate_status', 'share',
                     'project_id', 'size')

    instance_kwargs = {k: kwargs.get(k) for k in instance_keys if k in kwargs}
    snapshot_kwargs = {k: kwargs.get(k) for k in snapshot_keys if k in kwargs}

    aggregate_status = snapshot_kwargs.get(
        'aggregate_status', instance_kwargs.get(
            'status', constants.STATUS_CREATING))

    snapshot = {
        'id': 'fakesnapshotid',
        'share_name': 'fakename',
        'share_id': 'fakeid',
        'name': 'fakesnapshotname',
        'share_size': 1,
        'share_proto': 'fake_proto',
        'instance': {},
        'share': 'fake_share',
        'aggregate_status': aggregate_status,
        'project_id': 'fakeprojectid',
        'size': 1,
        'user_id': 'xyzzy',
    }
    snapshot.update(snapshot_kwargs)
    if create_instance:
        if 'instance_id' in instance_kwargs:
            instance_kwargs['id'] = instance_kwargs.pop('instance_id')
        snapshot['instance'] = fake_snapshot_instance(
            base_snapshot=snapshot, **instance_kwargs)
        snapshot['status'] = snapshot['instance']['status']
        snapshot['provider_location'] = (
            snapshot['instance']['provider_location']
        )
        snapshot['progress'] = snapshot['instance']['progress']
        snapshot['instances'] = snapshot['instance'],
    else:
        snapshot['status'] = constants.STATUS_AVAILABLE
        snapshot['progress'] = '0%'
        snapshot['provider_location'] = 'fake'
        snapshot.update(instance_kwargs)

    return db_fakes.FakeModel(snapshot)


def fake_snapshot_instance(base_snapshot=None, as_primitive=False, **kwargs):
    if base_snapshot is None:
        base_snapshot = fake_snapshot()
    snapshot_instance = {
        'id': 'fakesnapshotinstanceid',
        'snapshot_id': base_snapshot['id'],
        'status': constants.STATUS_CREATING,
        'progress': '0%',
        'provider_location': 'i_live_here_actually',
        'share_name': 'fakename',
        'share_id': 'fakeshareinstanceid',
        'share_instance': {
            'share_id': 'fakeshareid',
            'share_type_id': '1',
        },
        'share_instance_id': 'fakeshareinstanceid',
        'deleted': False,
        'updated_at': datetime.datetime(2016, 3, 21, 0, 5, 58),
        'created_at': datetime.datetime(2016, 3, 21, 0, 5, 58),
        'deleted_at': None,
        'share': fake_share(),
    }
    snapshot_instance.update(kwargs)
    if as_primitive:
        return snapshot_instance
    else:
        return db_fakes.FakeModel(snapshot_instance)


def expected_snapshot(version=None, id='fake_snapshot_id', **kwargs):
    api_major_version = 'v2' if version and version.startswith('2.') else 'v1'
    self_link = 'http://localhost/share/%s/fake/snapshots/%s' % (
        api_major_version, id)
    bookmark_link = 'http://localhost/share/fake/snapshots/%s' % id
    snapshot = {
        'id': id,
        'share_id': 'fakeshareid',
        'created_at': datetime.datetime(1, 1, 1, 1, 1, 1),
        'status': 'fakesnapstatus',
        'name': 'displaysnapname',
        'description': 'displaysnapdesc',
        'share_size': 1,
        'size': 1,
        'share_proto': 'fakesnapproto',
        'links': [
            {
                'href': self_link,
                'rel': 'self',
            },
            {
                'href': bookmark_link,
                'rel': 'bookmark',
            },
        ],
    }

    if version and (api_version.APIVersionRequest(version)
                    >= api_version.APIVersionRequest('2.17')):
        snapshot.update({
            'user_id': 'fakesnapuser',
            'project_id': 'fakesnapproject',
        })

    snapshot.update(kwargs)
    return {'snapshot': snapshot}


def search_opts(**kwargs):
    search_opts = {
        'name': 'fake_name',
        'status': 'fake_status',
        'share_id': 'fake_share_id',
        'sort_key': 'fake_sort_key',
        'sort_dir': 'fake_sort_dir',
        'offset': '1',
        'limit': '1',
    }
    search_opts.update(kwargs)
    return search_opts


def fake_access(**kwargs):
    access = {
        'id': 'fakeaccid',
        'access_type': 'ip',
        'access_to': '10.0.0.1',
        'access_level': 'rw',
        'state': 'active',
    }
    access.update(kwargs)
    return db_fakes.FakeModel(access)


def fake_replica(id=None, as_primitive=True, for_manager=False, **kwargs):
    replica = {
        'id': id or uuidutils.generate_uuid(),
        'share_id': 'f0e4bb5e-65f0-11e5-9d70-feff819cdc9f',
        'deleted': False,
        'host': 'openstack@BackendZ#PoolA',
        'status': 'available',
        'scheduled_at': datetime.datetime(2015, 8, 10, 0, 5, 58),
        'launched_at': datetime.datetime(2015, 8, 10, 0, 5, 58),
        'terminated_at': None,
        'replica_state': None,
        'availability_zone_id': 'f6e146d0-65f0-11e5-9d70-feff819cdc9f',
        'export_locations': [{'path': 'path1'}, {'path': 'path2'}],
        'share_network_id': '4ccd5318-65f1-11e5-9d70-feff819cdc9f',
        'share_server_id': '53099868-65f1-11e5-9d70-feff819cdc9f',
        'access_rules_status': constants.SHARE_INSTANCE_RULES_SYNCING,
    }
    if for_manager:
        replica.update({
            'user_id': None,
            'project_id': None,
            'share_type_id': None,
            'size': None,
            'display_name': None,
            'display_description': None,
            'replication_type': None,
            'snapshot_id': None,
            'share_proto': None,
            'is_public': None,
            'share_group_id': None,
            'source_share_group_snapshot_member_id': None,
            'availability_zone': 'fake_az',
        })
    replica.update(kwargs)
    if as_primitive:
        return replica
    else:
        return db_fakes.FakeModel(replica)


def fake_replica_request_spec(as_primitive=True, **kwargs):
    replica = fake_replica(id='9c0db763-a109-4862-b010-10f2bd395295')
    all_replica_hosts = ','.join(['fake_active_replica_host', replica['host']])
    request_spec = {
        'share_properties': fake_share(
            id='f0e4bb5e-65f0-11e5-9d70-feff819cdc9f'),
        'share_instance_properties': replica,
        'share_proto': 'nfs',
        'share_id': 'f0e4bb5e-65f0-11e5-9d70-feff819cdc9f',
        'snapshot_id': None,
        'share_type': 'fake_share_type',
        'share_group': None,
        'active_replica_host': 'fake_active_replica_host',
        'all_replica_hosts': all_replica_hosts,
    }
    request_spec.update(kwargs)
    if as_primitive:
        return request_spec
    else:
        return db_fakes.FakeModel(request_spec)


def fake_share_server_get():
    fake_share_server = {
        'status': constants.STATUS_ACTIVE,
        'updated_at': None,
        'host': 'fake_host',
        'share_network_subnet_id': 'fake_sn_id',
        'share_network_name': 'fake_sn_name',
        'project_id': 'fake_project_id',
        'id': 'fake_share_server_id',
        'backend_details': {
            'security_service_active_directory': '{"name": "fake_AD"}',
            'security_service_ldap': '{"name": "fake_LDAP"}',
            'security_service_kerberos': '{"name": "fake_kerberos"}',
        }
    }
    return fake_share_server



# Copyright (c) 2015 Clinton Knight.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
from oslo_utils import uuidutils
from webob import exc

from manila.api.openstack import api_version_request as api_version
from manila.api.v1 import scheduler_stats
from manila import context
from manila import exception
from manila import policy
from manila.scheduler import rpcapi
from manila.share import share_types
from manila import test
from manila.tests.api import fakes


FAKE_POOLS = [
    {
        'name': 'host1@backend1#pool1',
        'host': 'host1',
        'backend': 'backend1',
        'pool': 'pool1',
        'capabilities': {
            'updated': None,
            'total_capacity': 1024,
            'free_capacity': 100,
            'share_backend_name': 'pool1',
            'reserved_percentage': 0,
            'driver_version': '1.0.0',
            'storage_protocol': 'iSCSI',
            'qos': 'False',
        },
    },
    {
        'name': 'host1@backend1#pool2',
        'host': 'host1',
        'backend': 'backend1',
        'pool': 'pool2',
        'capabilities': {
            'updated': None,
            'total_capacity': 512,
            'free_capacity': 200,
            'share_backend_name': 'pool2',
            'reserved_percentage': 0,
            'driver_version': '1.0.1',
            'storage_protocol': 'iSER',
            'qos': 'True',
        },
    },
]


@ddt.ddt
class SchedulerStatsControllerTestCase(test.TestCase):
    def setUp(self):
        super(SchedulerStatsControllerTestCase, self).setUp()
        self.flags(host='fake')
        self.controller = scheduler_stats.SchedulerStatsController()
        self.resource_name = self.controller.resource_name
        self.ctxt = context.RequestContext('admin', 'fake', True)
        self.mock_policy_check = self.mock_object(
            policy, 'check_policy', mock.Mock(return_value=True))

    def test_pools_index(self):
        mock_get_pools = self.mock_object(rpcapi.SchedulerAPI,
                                          'get_pools',
                                          mock.Mock(return_value=FAKE_POOLS))
        req = fakes.HTTPRequest.blank('/v1/fake_project/scheduler_stats/pools')
        req.environ['manila.context'] = self.ctxt

        result = self.controller.pools_index(req)

        expected = {
            'pools': [
                {
                    'name': 'host1@backend1#pool1',
                    'host': 'host1',
                    'backend': 'backend1',
                    'pool': 'pool1',
                },
                {
                    'name': 'host1@backend1#pool2',
                    'host': 'host1',
                    'backend': 'backend1',
                    'pool': 'pool2',
                }
            ]
        }

        self.assertDictEqual(result, expected)
        mock_get_pools.assert_called_once_with(self.ctxt, filters={},
                                               cached=True)
        self.mock_policy_check.assert_called_once_with(
            self.ctxt, self.resource_name, 'index')

    @ddt.data(('index', False), ('detail', True))
    @ddt.unpack
    def test_pools_with_share_type_disabled(self, action, detail):
        mock_get_pools = self.mock_object(rpcapi.SchedulerAPI,
                                          'get_pools',
                                          mock.Mock(return_value=FAKE_POOLS))

        url = '/v1/fake_project/scheduler-stats/pools/%s' % action
        url += '?backend=back1&host=host1&pool=pool1'

        req = fakes.HTTPRequest.blank(url)
        req.environ['manila.context'] = self.ctxt

        expected_filters = {
            'host': 'host1',
            'pool': 'pool1',
            'backend': 'back1',
        }

        if detail:
            expected_result = {"pools": FAKE_POOLS}
        else:
            expected_result = {
                'pools': [
                    {
                        'name': 'host1@backend1#pool1',
                        'host': 'host1',
                        'backend': 'backend1',
                        'pool': 'pool1',
                    },
                    {
                        'name': 'host1@backend1#pool2',
                        'host': 'host1',
                        'backend': 'backend1',
                        'pool': 'pool2',
                    }
                ]
            }

        result = self.controller._pools(req, action, False)

        self.assertDictEqual(result, expected_result)
        mock_get_pools.assert_called_once_with(self.ctxt,
                                               filters=expected_filters,
                                               cached=True)

    @ddt.data(('index', False, True),
              ('index', False, False),
              ('detail', True, True),
              ('detail', True, False))
    @ddt.unpack
    def test_pools_with_share_type_enable(self, action, detail, uuid):
        mock_get_pools = self.mock_object(rpcapi.SchedulerAPI,
                                          'get_pools',
                                          mock.Mock(return_value=FAKE_POOLS))

        if uuid:
            share_type = uuidutils.generate_uuid()
        else:
            share_type = 'test_type'

        self.mock_object(
            share_types, 'get_share_type_by_name_or_id',
            mock.Mock(return_value={'extra_specs':
                                    {'snapshot_support': True}}))

        url = '/v1/fake_project/scheduler-stats/pools/%s' % action
        url += ('?backend=back1&host=host1&pool=pool1&share_type=%s'
                % share_type)

        req = fakes.HTTPRequest.blank(url)
        req.environ['manila.context'] = self.ctxt

        expected_filters = {
            'host': 'host1',
            'pool': 'pool1',
            'backend': 'back1',
            'capabilities': {
                'snapshot_support': True
            }
        }

        if detail:
            expected_result = {"pools": FAKE_POOLS}
        else:
            expected_result = {
                'pools': [
                    {
                        'name': 'host1@backend1#pool1',
                        'host': 'host1',
                        'backend': 'backend1',
                        'pool': 'pool1',
                    },
                    {
                        'name': 'host1@backend1#pool2',
                        'host': 'host1',
                        'backend': 'backend1',
                        'pool': 'pool2',
                    }
                ]
            }

        result = self.controller._pools(req, action, True)

        self.assertDictEqual(result, expected_result)
        mock_get_pools.assert_called_once_with(self.ctxt,
                                               filters=expected_filters,
                                               cached=True)

    @ddt.data('index', 'detail')
    def test_pools_with_share_type_not_found(self, action):
        url = '/v1/fake_project/scheduler-stats/pools/%s' % action
        url += '?backend=.%2A&host=host1&pool=pool%2A&share_type=fake_name_1'

        req = fakes.HTTPRequest.blank(url)

        self.assertRaises(exc.HTTPBadRequest,
                          self.controller._pools,
                          req, action, True)

    @ddt.data("1.0", "2.22", "2.23")
    def test_pools_index_with_filters(self, microversion):
        mock_get_pools = self.mock_object(rpcapi.SchedulerAPI,
                                          'get_pools',
                                          mock.Mock(return_value=FAKE_POOLS))
        self.mock_object(
            share_types, 'get_share_type_by_name',
            mock.Mock(return_value={'extra_specs':
                                    {'snapshot_support': True}}))

        url = '/v1/fake_project/scheduler-stats/pools/detail'
        url += '?backend=.%2A&host=host1&pool=pool%2A&share_type=test_type'

        req = fakes.HTTPRequest.blank(url, version=microversion)
        req.environ['manila.context'] = self.ctxt

        result = self.controller.pools_index(req)

        expected = {
            'pools': [
                {
                    'name': 'host1@backend1#pool1',
                    'host': 'host1',
                    'backend': 'backend1',
                    'pool': 'pool1',
                },
                {
                    'name': 'host1@backend1#pool2',
                    'host': 'host1',
                    'backend': 'backend1',
                    'pool': 'pool2',
                }
            ]
        }
        expected_filters = {
            'host': 'host1',
            'pool': 'pool*',
            'backend': '.*',
            'share_type': 'test_type',
        }
        if (api_version.APIVersionRequest(microversion) >=
                api_version.APIVersionRequest('2.23')):
            expected_filters.update(
                {'capabilities': {'snapshot_support': True}})
            expected_filters.pop('share_type', None)

        self.assertDictEqual(result, expected)
        mock_get_pools.assert_called_once_with(self.ctxt,
                                               filters=expected_filters,
                                               cached=True)
        self.mock_policy_check.assert_called_once_with(
            self.ctxt, self.resource_name, 'index')

    def test_get_pools_detail(self):
        mock_get_pools = self.mock_object(rpcapi.SchedulerAPI,
                                          'get_pools',
                                          mock.Mock(return_value=FAKE_POOLS))
        req = fakes.HTTPRequest.blank(
            '/v1/fake_project/scheduler_stats/pools/detail')
        req.environ['manila.context'] = self.ctxt

        result = self.controller.pools_detail(req)

        expected = {
            'pools': [
                {
                    'name': 'host1@backend1#pool1',
                    'host': 'host1',
                    'backend': 'backend1',
                    'pool': 'pool1',
                    'capabilities': {
                        'updated': None,
                        'total_capacity': 1024,
                        'free_capacity': 100,
                        'share_backend_name': 'pool1',
                        'reserved_percentage': 0,
                        'driver_version': '1.0.0',
                        'storage_protocol': 'iSCSI',
                        'qos': 'False',
                    },
                },
                {
                    'name': 'host1@backend1#pool2',
                    'host': 'host1',
                    'backend': 'backend1',
                    'pool': 'pool2',
                    'capabilities': {
                        'updated': None,
                        'total_capacity': 512,
                        'free_capacity': 200,
                        'share_backend_name': 'pool2',
                        'reserved_percentage': 0,
                        'driver_version': '1.0.1',
                        'storage_protocol': 'iSER',
                        'qos': 'True',
                    },
                },
            ],
        }

        self.assertDictEqual(expected, result)
        mock_get_pools.assert_called_once_with(self.ctxt, filters={},
                                               cached=True)
        self.mock_policy_check.assert_called_once_with(
            self.ctxt, self.resource_name, 'detail')

    @ddt.data('index', 'detail')
    def test_pools_forbidden(self, subresource):
        mock_get_pools = self.mock_object(
            rpcapi.SchedulerAPI, 'get_pools',
            mock.Mock(side_effect=exception.AdminRequired(
                "some traceback here")))
        path = '/v1/fake_project/scheduler_stats/pools'
        path = path + ('/%s' % subresource if subresource == 'detail' else '')
        req = fakes.HTTPRequest.blank(path)
        req.environ['manila.context'] = self.ctxt

        self.assertRaises(exc.HTTPForbidden,
                          getattr(self.controller, 'pools_%s' % subresource),
                          req)
        mock_get_pools.assert_called_once_with(self.ctxt,
                                               filters={},
                                               cached=True)


class SchedulerStatsTestCase(test.TestCase):

    def test_create_resource(self):
        result = scheduler_stats.create_resource()
        self.assertIsInstance(result.controller,
                              scheduler_stats.SchedulerStatsController)



# Copyright (c) 2015 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import re

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log

from manila import exception
from manila.i18n import _
from manila.share.drivers import service_instance
from manila.share.drivers.windows import windows_utils
from manila.share.drivers.windows import winrm_helper


CONF = cfg.CONF
LOG = log.getLogger(__name__)

windows_share_server_opts = [
    cfg.StrOpt(
        "winrm_cert_pem_path",
        default="~/.ssl/cert.pem",
        help="Path to the x509 certificate used for accessing the service "
             "instance."),
    cfg.StrOpt(
        "winrm_cert_key_pem_path",
        default="~/.ssl/key.pem",
        help="Path to the x509 certificate key."),
    cfg.BoolOpt(
        "winrm_use_cert_based_auth",
        default=False,
        help="Use x509 certificates in order to authenticate to the "
             "service instance.")
]

CONF = cfg.CONF
CONF.register_opts(windows_share_server_opts)


class WindowsServiceInstanceManager(service_instance.ServiceInstanceManager):
    """"Manages Windows Nova instances."""
    _INSTANCE_CONNECTION_PROTO = "WinRM"
    _CBS_INIT_RUN_PLUGIN_AFTER_REBOOT = 2
    _CBS_INIT_WINRM_PLUGIN = "ConfigWinRMListenerPlugin"
    _DEFAULT_MINIMUM_PASS_LENGTH = 6

    def __init__(self, driver_config=None, remote_execute=None):
        super(WindowsServiceInstanceManager, self).__init__(
            driver_config=driver_config)
        driver_config.append_config_values(windows_share_server_opts)

        self._use_cert_auth = self.get_config_option(
            "winrm_use_cert_based_auth")
        self._cert_pem_path = self.get_config_option(
            "winrm_cert_pem_path")
        self._cert_key_pem_path = self.get_config_option(
            "winrm_cert_key_pem_path")
        self._check_auth_mode()

        self._remote_execute = (remote_execute or
                                winrm_helper.WinRMHelper(
                                    configuration=driver_config).execute)
        self._windows_utils = windows_utils.WindowsUtils(
            remote_execute=self._remote_execute)

    def _check_auth_mode(self):
        if self._use_cert_auth:
            if not (os.path.exists(self._cert_pem_path) and
                    os.path.exists(self._cert_key_pem_path)):
                msg = _("Certificate based authentication was configured "
                        "but one or more certificates are missing.")
                raise exception.ServiceInstanceException(msg)
            LOG.debug("Using certificate based authentication for "
                      "service instances.")
        else:
            instance_password = self.get_config_option(
                "service_instance_password")
            if not self._check_password_complexity(instance_password):
                msg = _("The configured service instance password does not "
                        "match the minimum complexity requirements. "
                        "The password must contain at least %s characters. "
                        "Also, it must contain at least one digit, "
                        "one lower case and one upper case character.")
                raise exception.ServiceInstanceException(
                    msg % self._DEFAULT_MINIMUM_PASS_LENGTH)
            LOG.debug("Using password based authentication for "
                      "service instances.")

    def _get_auth_info(self):
        auth_info = {'use_cert_auth': self._use_cert_auth}
        if self._use_cert_auth:
            auth_info.update(cert_pem_path=self._cert_pem_path,
                             cert_key_pem_path=self._cert_key_pem_path)
        return auth_info

    def get_common_server(self):
        data = super(WindowsServiceInstanceManager, self).get_common_server()
        data['backend_details'].update(self._get_auth_info())
        return data

    def _get_new_instance_details(self, server):
        instance_details = super(WindowsServiceInstanceManager,
                                 self)._get_new_instance_details(server)
        instance_details.update(self._get_auth_info())
        return instance_details

    def _check_password_complexity(self, password):
        # Make sure that the Windows complexity requirements are met:
        # http://technet.microsoft.com/en-us/library/cc786468(v=ws.10).aspx
        if len(password) < self._DEFAULT_MINIMUM_PASS_LENGTH:
            return False
        for r in ("[a-z]", "[A-Z]", "[0-9]"):
            if not re.search(r, password):
                return False
        return True

    def _test_server_connection(self, server):
        try:
            self._remote_execute(server, "whoami", retry=False)
            LOG.debug("Service VM %s is available via WinRM",
                      server['ip'])
            return True
        except Exception as ex:
            LOG.debug("Server %(ip)s is not available via WinRM. "
                      "Exception: %(ex)s ",
                      dict(ip=server['ip'],
                           ex=ex))
            return False

    def _get_service_instance_create_kwargs(self):
        create_kwargs = {}
        if self._use_cert_auth:
            # At the moment, we pass the x509 certificate via user data.
            # We'll use keypairs instead as soon as the nova client will
            # support x509 certificates.
            with open(self._cert_pem_path, 'r') as f:
                cert_pem_data = f.read()
            create_kwargs['user_data'] = cert_pem_data
        else:
            # The admin password has to be specified via instance metadata in
            # order to be passed to the instance via the metadata service or
            # configdrive.
            admin_pass = self.get_config_option("service_instance_password")
            create_kwargs['meta'] = {'admin_pass': admin_pass}
        return create_kwargs

    def set_up_service_instance(self, context, network_info):
        instance_details = super(WindowsServiceInstanceManager,
                                 self).set_up_service_instance(context,
                                                               network_info)
        security_services = network_info['security_services']
        security_service = self.get_valid_security_service(security_services)
        if security_service:
            self._setup_security_service(instance_details, security_service)

        instance_details['joined_domain'] = bool(security_service)
        return instance_details

    def _setup_security_service(self, server, security_service):
        domain = security_service['domain']
        admin_username = security_service['user']
        admin_password = security_service['password']
        dns_ip = security_service['dns_ip']

        self._windows_utils.set_dns_client_search_list(server, [domain])

        if_index = self._windows_utils.get_interface_index_by_ip(server,
                                                                 server['ip'])
        self._windows_utils.set_dns_client_server_addresses(server,
                                                            if_index,
                                                            [dns_ip])
        # Joining an AD domain will alter the WinRM Listener configuration.
        # Cloudbase-init is required to be running on the Windows service
        # instance, so we re-enable the plugin configuring the WinRM listener.
        #
        # TODO(lpetrut): add a config option so that we may rely on the AD
        # group policies taking care of the WinRM configuration.
        self._run_cloudbase_init_plugin_after_reboot(
            server, plugin_name=self._CBS_INIT_WINRM_PLUGIN)
        self._join_domain(server, domain, admin_username, admin_password)

    def _join_domain(self, server, domain, admin_username, admin_password):
        # As the WinRM configuration may be altered and existing connections
        # closed, we may not be able to retrieve the result of this operation.
        # Instead, we'll ensure that the instance actually joined the domain
        # after the reboot.
        try:
            self._windows_utils.join_domain(server, domain, admin_username,
                                            admin_password)
        except processutils.ProcessExecutionError:
            raise
        except Exception as exc:
            LOG.debug("Unexpected error while attempting to join domain "
                      "%(domain)s. Verifying the result of the operation "
                      "after instance reboot. Exception: %(exc)s",
                      dict(domain=domain, exc=exc))
        # We reboot the service instance using the Compute API so that
        # we can wait for it to become active.
        self.reboot_server(server, soft_reboot=True)
        self.wait_for_instance_to_be_active(
            server['instance_id'],
            timeout=self.max_time_to_build_instance)
        if not self._check_server_availability(server):
            raise exception.ServiceInstanceException(
                _('%(conn_proto)s connection has not been '
                  'established to %(server)s in %(time)ss. Giving up.') % {
                      'conn_proto': self._INSTANCE_CONNECTION_PROTO,
                      'server': server['ip'],
                      'time': self.max_time_to_build_instance})

        current_domain = self._windows_utils.get_current_domain(server)
        if current_domain != domain:
            err_msg = _("Failed to join domain %(requested_domain)s. "
                        "Current domain: %(current_domain)s")
            raise exception.ServiceInstanceException(
                err_msg % dict(requested_domain=domain,
                               current_domain=current_domain))

    def get_valid_security_service(self, security_services):
        if not security_services:
            LOG.info("No security services provided.")
        elif len(security_services) > 1:
            LOG.warning("Multiple security services provided. Only one "
                        "security service of type 'active_directory' "
                        "is supported.")
        else:
            security_service = security_services[0]
            security_service_type = security_service['type']
            if security_service_type == 'active_directory':
                return security_service
            else:
                LOG.warning("Only security services of type "
                            "'active_directory' are supported. "
                            "Retrieved security "
                            "service type: %(sec_type)s.",
                            {'sec_type': security_service_type})
        return None

    def _run_cloudbase_init_plugin_after_reboot(self, server, plugin_name):
        cbs_init_reg_section = self._get_cbs_init_reg_section(server)
        plugin_key_path = "%(cbs_init_section)s\\%(instance_id)s\\Plugins" % {
            'cbs_init_section': cbs_init_reg_section,
            'instance_id': server['instance_id']
        }
        self._windows_utils.set_win_reg_value(
            server, path=plugin_key_path, key=plugin_name,
            value=self._CBS_INIT_RUN_PLUGIN_AFTER_REBOOT)

    def _get_cbs_init_reg_section(self, server):
        base_path = 'hklm:\\SOFTWARE'
        cbs_section = 'Cloudbase Solutions\\Cloudbase-Init'

        for upper_section in ('', 'Wow6432Node'):
            cbs_init_section = self._windows_utils.normalize_path(
                os.path.join(base_path, upper_section, cbs_section))
            try:
                self._windows_utils.get_win_reg_value(
                    server, path=cbs_init_section)
                return cbs_init_section
            except processutils.ProcessExecutionError as ex:
                # The exit code will always be '1' in case of errors, so the
                # only way to determine the error type is checking stderr.
                if 'Cannot find path' in ex.stderr:
                    continue
                else:
                    raise
        raise exception.ServiceInstanceException(
            _("Could not retrieve Cloudbase Init registry section"))



# Copyright 2010 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils
import webob
import webob.dec
import webob.exc

from manila.api.middleware import fault
from manila.api.openstack import wsgi
from manila import exception
from manila import test


class TestFaults(test.TestCase):
    """Tests covering `manila.api.openstack.faults:Fault` class."""

    def _prepare_xml(self, xml_string):
        """Remove characters from string which hinder XML equality testing."""
        xml_string = xml_string.replace("  ", "")
        xml_string = xml_string.replace("\n", "")
        xml_string = xml_string.replace("\t", "")
        return xml_string

    def test_400_fault_json(self):
        """Test fault serialized to JSON via file-extension and/or header."""
        requests = [
            webob.Request.blank('/.json'),
            webob.Request.blank('/', headers={"Accept": "application/json"}),
        ]

        for request in requests:
            fault = wsgi.Fault(webob.exc.HTTPBadRequest(explanation='scram'))
            response = request.get_response(fault)

            expected = {
                "badRequest": {
                    "message": "scram",
                    "code": 400,
                },
            }
            actual = jsonutils.loads(response.body)

            self.assertEqual("application/json", response.content_type)
            self.assertEqual(expected, actual)

    def test_413_fault_json(self):
        """Test fault serialized to JSON via file-extension and/or header."""
        requests = [
            webob.Request.blank('/.json'),
            webob.Request.blank('/', headers={"Accept": "application/json"}),
        ]

        for request in requests:
            exc = webob.exc.HTTPRequestEntityTooLarge
            fault = wsgi.Fault(exc(explanation='sorry',
                                   headers={'Retry-After': 4}))
            response = request.get_response(fault)

            expected = {
                "overLimit": {
                    "message": "sorry",
                    "code": 413,
                    "retryAfter": '4',
                },
            }
            actual = jsonutils.loads(response.body)

            self.assertEqual("application/json", response.content_type)
            self.assertEqual(expected, actual)

    def test_raise(self):
        """Ensure the ability to raise :class:`Fault` in WSGI-ified methods."""
        @webob.dec.wsgify
        def raiser(req):
            raise wsgi.Fault(webob.exc.HTTPNotFound(explanation='whut?'))

        req = webob.Request.blank('/.json')
        resp = req.get_response(raiser)
        self.assertEqual("application/json", resp.content_type)
        self.assertEqual(404, resp.status_int)
        self.assertIn('whut?'.encode("utf-8"), resp.body)

    def test_raise_403(self):
        """Ensure the ability to raise :class:`Fault` in WSGI-ified methods."""
        @webob.dec.wsgify
        def raiser(req):
            raise wsgi.Fault(webob.exc.HTTPForbidden(explanation='whut?'))

        req = webob.Request.blank('/.json')
        resp = req.get_response(raiser)
        self.assertEqual("application/json", resp.content_type)
        self.assertEqual(403, resp.status_int)
        self.assertNotIn('resizeNotAllowed'.encode("utf-8"), resp.body)
        self.assertIn('forbidden'.encode("utf-8"), resp.body)

    def test_fault_has_status_int(self):
        """Ensure the status_int is set correctly on faults."""
        fault = wsgi.Fault(webob.exc.HTTPBadRequest(explanation='what?'))
        self.assertEqual(400, fault.status_int)


class ExceptionTest(test.TestCase):

    def _wsgi_app(self, inner_app):
        return fault.FaultWrapper(inner_app)

    def _do_test_exception_safety_reflected_in_faults(self, expose):
        class ExceptionWithSafety(exception.ManilaException):
            safe = expose

        @webob.dec.wsgify
        def fail(req):
            raise ExceptionWithSafety('some explanation')

        api = self._wsgi_app(fail)
        resp = webob.Request.blank('/').get_response(api)
        self.assertIn('{"computeFault', str(resp.body), resp.body)
        expected = ('ExceptionWithSafety: some explanation' if expose else
                    'The server has either erred or is incapable '
                    'of performing the requested operation.')
        self.assertIn(expected, str(resp.body), resp.body)
        self.assertEqual(500, resp.status_int, resp.body)

    def test_safe_exceptions_are_described_in_faults(self):
        self._do_test_exception_safety_reflected_in_faults(True)

    def test_unsafe_exceptions_are_not_described_in_faults(self):
        self._do_test_exception_safety_reflected_in_faults(False)

    def _do_test_exception_mapping(self, exception_type, msg):
        @webob.dec.wsgify
        def fail(req):
            raise exception_type(msg)

        api = self._wsgi_app(fail)
        resp = webob.Request.blank('/').get_response(api)
        self.assertIn(msg, str(resp.body), resp.body)
        self.assertEqual(exception_type.code, resp.status_int, resp.body)

        if hasattr(exception_type, 'headers'):
            for (key, value) in exception_type.headers.items():
                self.assertIn(key, resp.headers)
                self.assertEqual(value, resp.headers[key])

    def test_quota_error_mapping(self):
        self._do_test_exception_mapping(exception.QuotaError, 'too many used')

    def test_non_manila_notfound_exception_mapping(self):
        class ExceptionWithCode(Exception):
            code = 404

        self._do_test_exception_mapping(ExceptionWithCode,
                                        'NotFound')

    def test_non_manila_exception_mapping(self):
        class ExceptionWithCode(Exception):
            code = 417

        self._do_test_exception_mapping(ExceptionWithCode,
                                        'Expectation failed')

    def test_exception_with_none_code_throws_500(self):
        class ExceptionWithNoneCode(Exception):
            code = None

        @webob.dec.wsgify
        def fail(req):
            raise ExceptionWithNoneCode()

        api = self._wsgi_app(fail)
        resp = webob.Request.blank('/').get_response(api)
        self.assertEqual(500, resp.status_int)

    def test_validate_request_unicode_decode_fault(self):
        @webob.dec.wsgify
        def unicode_error(req):
            raise UnicodeDecodeError("ascii", "test".encode(), 0, 1, "bad")

        api = self._wsgi_app(unicode_error)
        resp = webob.Request.blank('/test?foo=%88').get_response(api)
        self.assertEqual(400, resp.status_int)



# Copyright (c) 2015 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import os

from oslo_log import log

from manila.common import constants
from manila import exception
from manila.share.drivers import helpers
from manila.share.drivers.windows import windows_utils

LOG = log.getLogger(__name__)


class WindowsSMBHelper(helpers.CIFSHelperBase):
    _SHARE_ACCESS_RIGHT_MAP = {
        constants.ACCESS_LEVEL_RW: "Change",
        constants.ACCESS_LEVEL_RO: "Read"}

    _NULL_SID = "S-1-0-0"

    _WIN_ACL_ALLOW = 0
    _WIN_ACL_DENY = 1

    _WIN_ACCESS_RIGHT_FULL = 0
    _WIN_ACCESS_RIGHT_CHANGE = 1
    _WIN_ACCESS_RIGHT_READ = 2
    _WIN_ACCESS_RIGHT_CUSTOM = 3

    _ACCESS_LEVEL_CUSTOM = 'custom'

    _WIN_ACL_MAP = {
        _WIN_ACCESS_RIGHT_CHANGE: constants.ACCESS_LEVEL_RW,
        _WIN_ACCESS_RIGHT_FULL: constants.ACCESS_LEVEL_RW,
        _WIN_ACCESS_RIGHT_READ: constants.ACCESS_LEVEL_RO,
        _WIN_ACCESS_RIGHT_CUSTOM: _ACCESS_LEVEL_CUSTOM,
    }

    _SUPPORTED_ACCESS_LEVELS = (constants.ACCESS_LEVEL_RO,
                                constants.ACCESS_LEVEL_RW)
    _SUPPORTED_ACCESS_TYPES = ('user', )

    def __init__(self, remote_execute, configuration):
        self._remote_exec = remote_execute
        self.configuration = configuration
        self._windows_utils = windows_utils.WindowsUtils(
            remote_execute=remote_execute)

    def init_helper(self, server):
        self._remote_exec(server, "Get-SmbShare")

    def create_exports(self, server, share_name, recreate=False):
        export_location = '\\\\%s\\%s' % (server['public_address'],
                                          share_name)
        if not self._share_exists(server, share_name):
            share_path = self._windows_utils.normalize_path(
                os.path.join(self.configuration.share_mount_path,
                             share_name))
            # If no access rules are requested, 'Everyone' will have read
            # access, by default. We set read access for the 'NULL SID' in
            # order to avoid this.
            cmd = ['New-SmbShare', '-Name', share_name, '-Path', share_path,
                   '-ReadAccess', "*%s" % self._NULL_SID]
            self._remote_exec(server, cmd)
        else:
            LOG.info("Skipping creating export %s as it already exists.",
                     share_name)
        return self.get_exports_for_share(server, export_location)

    def remove_exports(self, server, share_name):
        if self._share_exists(server, share_name):
            cmd = ['Remove-SmbShare', '-Name', share_name, "-Force"]
            self._remote_exec(server, cmd)
        else:
            LOG.debug("Skipping removing export %s as it does not exist.",
                      share_name)

    def _get_volume_path_by_share_name(self, server, share_name):
        share_path = self._get_share_path_by_name(server, share_name)
        volume_path = self._windows_utils.get_volume_path_by_mount_path(
            server, share_path)
        return volume_path

    def _get_acls(self, server, share_name):
        cmd = ('Get-SmbShareAccess -Name %(share_name)s | '
               'Select-Object @("Name", "AccountName", '
               '"AccessControlType", "AccessRight") | '
               'ConvertTo-JSON -Compress' % {'share_name': share_name})
        (out, err) = self._remote_exec(server, cmd)

        if not out.strip():
            return []

        raw_acls = json.loads(out)
        if isinstance(raw_acls, dict):
            return [raw_acls]
        return raw_acls

    def get_access_rules(self, server, share_name):
        raw_acls = self._get_acls(server, share_name)
        acls = []

        for raw_acl in raw_acls:
            access_to = raw_acl['AccountName']
            access_right = raw_acl['AccessRight']
            access_level = self._WIN_ACL_MAP[access_right]
            access_allow = raw_acl["AccessControlType"] == self._WIN_ACL_ALLOW

            if not access_allow:
                if access_to.lower() == 'everyone' and len(raw_acls) == 1:
                    LOG.debug("No access rules are set yet for share %s",
                              share_name)
                else:
                    LOG.warning(
                        "Found explicit deny ACE rule that was not "
                        "created by Manila and will be ignored: %s",
                        raw_acl)
                continue
            if access_level == self._ACCESS_LEVEL_CUSTOM:
                LOG.warning(
                    "Found 'custom' ACE rule that will be ignored: %s",
                    raw_acl)
                continue
            elif access_right == self._WIN_ACCESS_RIGHT_FULL:
                LOG.warning(
                    "Account '%(access_to)s' was given full access "
                    "right on share %(share_name)s. Manila only "
                    "grants 'change' access.",
                    {'access_to': access_to,
                     'share_name': share_name})

            acl = {
                'access_to': access_to,
                'access_level': access_level,
                'access_type': 'user',
            }
            acls.append(acl)
        return acls

    def _grant_share_access(self, server, share_name, access_level, access_to):
        access_right = self._SHARE_ACCESS_RIGHT_MAP[access_level]
        cmd = ["Grant-SmbShareAccess", "-Name", share_name,
               "-AccessRight", access_right,
               "-AccountName", "'%s'" % access_to, "-Force"]
        self._remote_exec(server, cmd)
        self._refresh_acl(server, share_name)
        LOG.info("Granted %(access_level)s access to '%(access_to)s' "
                 "on share %(share_name)s",
                 {'access_level': access_level,
                  'access_to': access_to,
                  'share_name': share_name})

    def _refresh_acl(self, server, share_name):
        cmd = ['Set-SmbPathAcl', '-ShareName', share_name]
        self._remote_exec(server, cmd)

    def _revoke_share_access(self, server, share_name, access_to):
        cmd = ['Revoke-SmbShareAccess', '-Name', share_name,
               '-AccountName', '"%s"' % access_to, '-Force']
        self._remote_exec(server, cmd)
        self._refresh_acl(server, share_name)
        LOG.info("Revoked access to '%(access_to)s' "
                 "on share %(share_name)s",
                 {'access_to': access_to,
                  'share_name': share_name})

    def update_access(self, server, share_name, access_rules, add_rules,
                      delete_rules):
        self.validate_access_rules(
            access_rules + add_rules,
            self._SUPPORTED_ACCESS_TYPES,
            self._SUPPORTED_ACCESS_LEVELS)

        if not (add_rules or delete_rules):
            existing_rules = self.get_access_rules(server, share_name)
            add_rules, delete_rules = self._get_rule_updates(
                existing_rules=existing_rules,
                requested_rules=access_rules)
            LOG.debug(("Missing rules: %(add_rules)s, "
                       "superfluous rules: %(delete_rules)s"),
                      {'add_rules': add_rules,
                       'delete_rules': delete_rules})

        # Some rules may have changed, so we'll
        # treat the deleted rules first.
        for deleted_rule in delete_rules:
            try:
                self.validate_access_rules(
                    [deleted_rule],
                    self._SUPPORTED_ACCESS_TYPES,
                    self._SUPPORTED_ACCESS_LEVELS)
            except (exception.InvalidShareAccess,
                    exception.InvalidShareAccessLevel):
                # This check will allow invalid rules to be deleted.
                LOG.warning(
                    "Unsupported access level %(level)s or access type "
                    "%(type)s, skipping removal of access rule to "
                    "%(to)s.", {'level': deleted_rule['access_level'],
                                'type': deleted_rule['access_type'],
                                'to': deleted_rule['access_to']})
                continue
            self._revoke_share_access(server, share_name,
                                      deleted_rule['access_to'])

        for added_rule in add_rules:
            self._grant_share_access(server, share_name,
                                     added_rule['access_level'],
                                     added_rule['access_to'])

    def _subtract_access_rules(self, access_rules, subtracted_rules):
        # Account names are case insensitive on Windows.
        filter_rules = lambda rules: [  # noqa: E731
            {'access_to': access_rule['access_to'].lower(),
             'access_level': access_rule['access_level'],
             'access_type': access_rule['access_type']}
            for access_rule in rules]

        return [rule for rule in filter_rules(access_rules)
                if rule not in filter_rules(subtracted_rules)]

    def _get_rule_updates(self, existing_rules, requested_rules):
        added_rules = self._subtract_access_rules(requested_rules,
                                                  existing_rules)
        deleted_rules = self._subtract_access_rules(existing_rules,
                                                    requested_rules)
        return added_rules, deleted_rules

    def _get_share_name(self, export_location):
        return self._windows_utils.normalize_path(
            export_location).split('\\')[-1]

    def _get_export_location_template(self, old_export_location):
        share_name = self._get_share_name(old_export_location)
        return '\\\\%s' + ('\\%s' % share_name)

    def _get_share_path_by_name(self, server, share_name,
                                ignore_missing=False):
        cmd = ('Get-SmbShare -Name %s | '
               'Select-Object -ExpandProperty Path' % share_name)

        check_exit_code = not ignore_missing
        (share_path, err) = self._remote_exec(server, cmd,
                                              check_exit_code=check_exit_code)
        return share_path.strip() if share_path else None

    def get_share_path_by_export_location(self, server, export_location):
        share_name = self._get_share_name(export_location)
        return self._get_share_path_by_name(server, share_name)

    def _share_exists(self, server, share_name):
        share_path = self._get_share_path_by_name(server, share_name,
                                                  ignore_missing=True)
        return bool(share_path)



# Copyright 2019 Nexenta by DDN, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
:mod:`nexenta.options` -- Contains configuration options for Nexenta drivers.
=============================================================================

.. automodule:: nexenta.options
"""

from oslo_config import cfg

nexenta_connection_opts = [
    cfg.ListOpt('nexenta_rest_addresses',
                help='One or more comma delimited IP addresses for management '
                     'communication with NexentaStor appliance.'),
    cfg.IntOpt('nexenta_rest_port',
               default=8443,
               help='Port to connect to Nexenta REST API server.'),
    cfg.StrOpt('nexenta_rest_protocol',
               default='auto',
               choices=['http', 'https', 'auto'],
               help='Use http or https for REST connection (default auto).'),
    cfg.BoolOpt('nexenta_use_https',
                default=True,
                help='Use HTTP secure protocol for NexentaStor '
                     'management REST API connections'),
    cfg.StrOpt('nexenta_user',
               default='admin',
               help='User name to connect to Nexenta SA.',
               required=True),
    cfg.StrOpt('nexenta_password',
               help='Password to connect to Nexenta SA.',
               required=True,
               secret=True),
    cfg.StrOpt('nexenta_volume',
               default='volume1',
               help='Volume name on NexentaStor.'),
    cfg.StrOpt('nexenta_pool',
               default='pool1',
               required=True,
               help='Pool name on NexentaStor.'),
    cfg.BoolOpt('nexenta_nfs',
                default=True,
                help='Defines whether share over NFS is enabled.'),
    cfg.BoolOpt('nexenta_ssl_cert_verify',
                default=False,
                help='Defines whether the driver should check ssl cert.'),
    cfg.FloatOpt('nexenta_rest_connect_timeout',
                 default=30,
                 help='Specifies the time limit (in seconds), within '
                      'which the connection to NexentaStor management '
                      'REST API server must be established'),
    cfg.FloatOpt('nexenta_rest_read_timeout',
                 default=300,
                 help='Specifies the time limit (in seconds), '
                      'within which NexentaStor management '
                      'REST API server must send a response'),
    cfg.FloatOpt('nexenta_rest_backoff_factor',
                 default=1,
                 help='Specifies the backoff factor to apply '
                      'between connection attempts to NexentaStor '
                      'management REST API server'),
    cfg.IntOpt('nexenta_rest_retry_count',
               default=5,
               help='Specifies the number of times to repeat NexentaStor '
                    'management REST API call in case of connection errors '
                    'and NexentaStor appliance EBUSY or ENOENT errors'),
]

nexenta_nfs_opts = [
    cfg.HostAddressOpt('nexenta_nas_host',
                       help='Data IP address of Nexenta storage appliance.',
                       required=True),
    cfg.StrOpt('nexenta_mount_point_base',
               default='$state_path/mnt',
               help='Base directory that contains NFS share mount points.'),
]

nexenta_dataset_opts = [
    cfg.StrOpt('nexenta_nfs_share',
               default='nfs_share',
               help='Parent filesystem where all the shares will be created. '
                    'This parameter is only used by NexentaStor4 driver.'),
    cfg.StrOpt('nexenta_share_name_prefix',
               help='Nexenta share name prefix.',
               default='share-'),
    cfg.StrOpt('nexenta_folder',
               default='folder',
               required=True,
               help='Parent folder on NexentaStor.'),
    cfg.StrOpt('nexenta_dataset_compression',
               default='on',
               choices=['on', 'off', 'gzip', 'gzip-1', 'gzip-2', 'gzip-3',
                        'gzip-4', 'gzip-5', 'gzip-6', 'gzip-7', 'gzip-8',
                        'gzip-9', 'lzjb', 'zle', 'lz4'],
               help='Compression value for new ZFS folders.'),
    cfg.StrOpt('nexenta_dataset_dedupe',
               default='off',
               choices=['on', 'off', 'sha256', 'verify', 'sha256, verify'],
               help='Deduplication value for new ZFS folders. '
                    'Only used by NexentaStor4 driver.'),
    cfg.BoolOpt('nexenta_thin_provisioning',
                default=True,
                help=('If True shares will not be space guaranteed and '
                      'overprovisioning will be enabled.')),
    cfg.IntOpt('nexenta_dataset_record_size',
               default=131072,
               help='Specifies a suggested block size in for files in a file '
                    'system. (bytes)'),
]



# Copyright 2011 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import webob.exc

from manila.api import extensions
from manila.api.openstack import wsgi


class FoxInSocksController(object):

    def index(self, req):
        return "Try to say this Mr. Knox, sir..."


class FoxInSocksServerControllerExtension(wsgi.Controller):
    @wsgi.action('add_tweedle')
    def _add_tweedle(self, req, id, body):

        return "Tweedle Beetle Added."

    @wsgi.action('delete_tweedle')
    def _delete_tweedle(self, req, id, body):

        return "Tweedle Beetle Deleted."

    @wsgi.action('fail')
    def _fail(self, req, id, body):

        raise webob.exc.HTTPBadRequest(explanation='Tweedle fail')


class FoxInSocksFlavorGooseControllerExtension(wsgi.Controller):
    @wsgi.extends
    def show(self, req, resp_obj, id):
        # NOTE: This only handles JSON responses.
        # You can use content type header to test for XML.
        resp_obj.obj['flavor']['googoose'] = req.GET.get('chewing')


class FoxInSocksFlavorBandsControllerExtension(wsgi.Controller):
    @wsgi.extends
    def show(self, req, resp_obj, id):
        # NOTE: This only handles JSON responses.
        # You can use content type header to test for XML.
        resp_obj.obj['big_bands'] = 'Pig Bands!'


class Foxinsocks(extensions.ExtensionDescriptor):
    """The Fox In Socks Extension."""

    name = "Fox In Socks"
    alias = "FOXNSOX"
    namespace = "http://www.fox.in.socks/api/ext/pie/v1.0"
    updated = "2011-01-22T13:25:27-06:00"

    def __init__(self, ext_mgr):
        ext_mgr.register(self)

    def get_resources(self):
        resources = []
        resource = extensions.ResourceExtension('foxnsocks',
                                                FoxInSocksController())
        resources.append(resource)
        return resources

    def get_controller_extensions(self):
        extension_list = []

        extension_set = [
            (FoxInSocksServerControllerExtension, 'servers'),
            (FoxInSocksFlavorGooseControllerExtension, 'flavors'),
            (FoxInSocksFlavorBandsControllerExtension, 'flavors'), ]
        for klass, collection in extension_set:
            controller = klass()
            ext = extensions.ControllerExtension(self, collection, controller)
            extension_list.append(ext)

        return extension_list



# Copyright 2018 SAP SE
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Add ou to security service

Revision ID: 0274d20c560f
Revises: 4a482571410f
Create Date: 2017-05-19 17:27:30.274440

"""

# revision identifiers, used by Alembic.
revision = '0274d20c560f'
down_revision = '4a482571410f'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        'security_services',
        sa.Column('ou', sa.String(255), nullable=True))


def downgrade():
    op.drop_column('security_services', 'ou')



# Copyright (c) 2016 Goutham Pacha Ravi
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log

from manila.scheduler.filters import base_host

LOG = log.getLogger(__name__)


class ShareReplicationFilter(base_host.BaseHostFilter):
    """ShareReplicationFilter filters hosts based on replication support."""

    def host_passes(self, host_state, filter_properties):
        """Return True if 'active' replica's host can replicate with host.

        Design of this filter:

            - Share replication is symmetric. All backends that can
              replicate between each other must share the same
              'replication_domain'.
            - For scheduling a share that can be replicated in the future,
              this filter checks for 'replication_domain' capability.
            - For scheduling a replica, it checks for the
              'replication_domain' compatibility.

        """
        active_replica_host = filter_properties.get('request_spec', {}).get(
            'active_replica_host')
        existing_replica_hosts = filter_properties.get('request_spec', {}).get(
            'all_replica_hosts', '').split(',')
        replication_type = filter_properties.get('resource_type', {}).get(
            'extra_specs', {}).get('replication_type')
        active_replica_replication_domain = filter_properties.get(
            'replication_domain')
        host_replication_domain = host_state.replication_domain

        if replication_type is None:
            # NOTE(gouthamr): You're probably not creating a replicated
            # share or a replica, then this host obviously passes.
            return True
        elif host_replication_domain is None:
            msg = "Replication is not enabled on host %s."
            LOG.debug(msg, host_state.host)
            return False
        elif active_replica_host is None:
            # 'replication_type' filtering will be handled by the
            # capabilities filter, since it is a share-type extra-spec.
            return True

        # Scheduler filtering by replication_domain for a replica
        if active_replica_replication_domain != host_replication_domain:
            msg = ("The replication domain of Host %(host)s is "
                   "'%(host_domain)s' and it does not match the replication "
                   "domain of the 'active' replica's host: "
                   "%(active_replica_host)s, which is '%(arh_domain)s'. ")
            kwargs = {
                "host": host_state.host,
                "host_domain": host_replication_domain,
                "active_replica_host": active_replica_host,
                "arh_domain": active_replica_replication_domain,
            }
            LOG.debug(msg, kwargs)
            return False

        # Check host string for already created replicas
        if host_state.host in existing_replica_hosts:
            msg = ("Skipping host %s since it already hosts a replica for "
                   "this share.")
            LOG.debug(msg, host_state.host)
            return False

        return True



#    Copyright 2014 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools
import sys
import textwrap
from unittest import mock

import ddt
import pycodestyle

from manila import test
from manila.tests.hacking import checks


@ddt.ddt
class HackingTestCase(test.TestCase):
    """Hacking test cases

    This class tests the hacking checks in manila.tests.hacking.checks by
    passing strings to the check methods like the pep8/flake8 parser would.
    The parser loops over each line in the file and then passes the
    parameters to the check method. The parameter names in the check method
    dictate what type of object is passed to the check method. The
    parameter types are::

        logical_line: A processed line with the following modifications:
            - Multi-line statements converted to a single line.
            - Stripped left and right.
            - Contents of strings replaced with "xxx" of same length.
            - Comments removed.
        physical_line: Raw line of text from the input file.
        lines: a list of the raw lines from the input file
        tokens: the tokens that contribute to this logical line
        line_number: line number in the input file
        total_lines: number of lines in the input file
        blank_lines: blank lines before this one
        indent_char: indentation character in this file (" " or "\t")
        indent_level: indentation (with tabs expanded to multiples of 8)
        previous_indent_level: indentation on previous line
        previous_logical: previous logical line
        filename: Path of the file being run through pep8

    When running a test on a check method the return will be False/None if
    there is no violation in the sample input. If there is an error a tuple is
    returned with a position in the line, and a message. So to check the result
    just assertTrue if the check is expected to fail and assertFalse if it
    should pass.
    """

    @ddt.data(*itertools.product(
        ('', '_', '_LE', '_LI', '_LW'),
        ('audit', 'debug', 'error', 'info', 'warn', 'warning', 'critical',
         'exception',)))
    @ddt.unpack
    def test_no_translate_logs(self, log_marker, log_method):
        code = "LOG.{0}({1}('foo'))".format(log_method, log_marker)

        if log_marker:
            self.assertEqual(1, len(list(checks.no_translate_logs(code))))
        else:
            self.assertEqual(0, len(list(checks.no_translate_logs(code))))

    def test_check_explicit_underscore_import(self):
        self.assertEqual(1, len(list(checks.check_explicit_underscore_import(
            "LOG.info(_('My info message'))",
            "manila/tests/other_files.py"))))
        self.assertEqual(1, len(list(checks.check_explicit_underscore_import(
            "msg = _('My message')",
            "manila/tests/other_files.py"))))
        self.assertEqual(0, len(list(checks.check_explicit_underscore_import(
            "from manila.i18n import _",
            "manila/tests/other_files.py"))))
        self.assertEqual(0, len(list(checks.check_explicit_underscore_import(
            "LOG.info(_('My info message'))",
            "manila/tests/other_files.py"))))
        self.assertEqual(0, len(list(checks.check_explicit_underscore_import(
            "msg = _('My message')",
            "manila/tests/other_files.py"))))
        self.assertEqual(0, len(list(checks.check_explicit_underscore_import(
            "from manila.i18n import _LE, _, _LW",
            "manila/tests/other_files2.py"))))
        self.assertEqual(0, len(list(checks.check_explicit_underscore_import(
            "msg = _('My message')",
            "manila/tests/other_files2.py"))))
        self.assertEqual(0, len(list(checks.check_explicit_underscore_import(
            "_ = translations.ugettext",
            "manila/tests/other_files3.py"))))
        self.assertEqual(0, len(list(checks.check_explicit_underscore_import(
            "msg = _('My message')",
            "manila/tests/other_files3.py"))))
        # Complete code coverage by falling through all checks
        self.assertEqual(0, len(list(checks.check_explicit_underscore_import(
            "LOG.info('My info message')",
            "manila.tests.unit/other_files4.py"))))
        self.assertEqual(0, len(list(checks.check_explicit_underscore_import(
            "from manila.i18n import _LW",
            "manila.tests.unit/other_files5.py"))))
        self.assertEqual(1, len(list(checks.check_explicit_underscore_import(
            "msg = _('My message')",
            "manila.tests.unit/other_files5.py"))))

    # We are patching pep8 so that only the check under test is actually
    # installed.
    @mock.patch('pycodestyle._checks',
                {'physical_line': {}, 'logical_line': {}, 'tree': {}})
    def _run_check(self, code, checker, filename=None):
        pycodestyle.register_check(checker)

        lines = textwrap.dedent(code).strip().splitlines(True)

        checker = pycodestyle.Checker(filename=filename, lines=lines)
        checker.check_all()
        checker.report._deferred_print.sort()
        return checker.report._deferred_print

    def _assert_has_errors(self, code, checker, expected_errors=None,
                           filename=None):
        actual_errors = [e[:3] for e in
                         self._run_check(code, checker, filename)]
        self.assertEqual(expected_errors or [], actual_errors)

    def _assert_has_no_errors(self, code, checker, filename=None):
        self._assert_has_errors(code, checker, filename=filename)

    def test_logging_format_no_tuple_arguments(self):
        checker = checks.CheckLoggingFormatArgs
        code = """
               import logging
               LOG = logging.getLogger()
               LOG.info("Message without a second argument.")
               LOG.critical("Message with %s arguments.", 'two')
               LOG.debug("Volume %s caught fire and is at %d degrees C and"
                         " climbing.", 'volume1', 500)
               """
        self._assert_has_no_errors(code, checker)

    @ddt.data(*checks.CheckLoggingFormatArgs.LOG_METHODS)
    def test_logging_with_tuple_argument(self, log_method):
        checker = checks.CheckLoggingFormatArgs
        code = """
               import logging
               LOG = logging.getLogger()
               LOG.{0}("Volume %s caught fire and is at %d degrees C and "
                      "climbing.", ('volume1', 500))
               """
        self._assert_has_errors(code.format(log_method), checker,
                                expected_errors=[(4, mock.ANY, 'M310')])

    def test_trans_add(self):

        checker = checks.CheckForTransAdd
        code = """
               def fake_tran(msg):
                   return msg


               _ = fake_tran
               _LI = _
               _LW = _
               _LE = _
               _LC = _


               def f(a, b):
                   msg = _('test') + 'add me'
                   msg = _LI('test') + 'add me'
                   msg = _LW('test') + 'add me'
                   msg = _LE('test') + 'add me'
                   msg = _LC('test') + 'add me'
                   msg = 'add to me' + _('test')
                   return msg
               """

        # Python 3.4.0 introduced a change to the column calculation during AST
        # parsing. This was reversed in Python 3.4.3, hence the version-based
        # expected value calculation. See #1499743 for more background.
        if sys.version_info < (3, 4, 0) or sys.version_info >= (3, 4, 3):
            errors = [(13, 10, 'M326'), (14, 10, 'M326'), (15, 10, 'M326'),
                      (16, 10, 'M326'), (17, 10, 'M326'), (18, 24, 'M326')]
        else:
            errors = [(13, 11, 'M326'), (14, 13, 'M326'), (15, 13, 'M326'),
                      (16, 13, 'M326'), (17, 13, 'M326'), (18, 25, 'M326')]
        self._assert_has_errors(code, checker, expected_errors=errors)

        code = """
               def f(a, b):
                   msg = 'test' + 'add me'
                   return msg
               """
        errors = []
        self._assert_has_errors(code, checker, expected_errors=errors)

    def test_dict_constructor_with_list_copy(self):
        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "    dict([(i, connect_info[i])"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "    attrs = dict([(k, _from_json(v))"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "        type_names = dict((value, key) for key, value in"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "   dict((value, key) for key, value in"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "foo(param=dict((k, v) for k, v in bar.items()))"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            " dict([[i,i] for i in range(3)])"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "  dd = dict([i,i] for i in range(3))"))))

        self.assertEqual(0, len(list(checks.dict_constructor_with_list_copy(
            "        create_kwargs = dict(snapshot=snapshot,"))))

        self.assertEqual(0, len(list(checks.dict_constructor_with_list_copy(
            "      self._render_dict(xml, data_el, data.__dict__)"))))

    def test_no_xrange(self):
        self.assertEqual(1, len(list(checks.no_xrange("xrange(45)"))))

        self.assertEqual(0, len(list(checks.no_xrange("range(45)"))))

    def test_validate_assertTrue(self):
        test_value = True
        self.assertEqual(0, len(list(checks.validate_assertTrue(
            "assertTrue(True)"))))
        self.assertEqual(1, len(list(checks.validate_assertTrue(
            "assertEqual(True, %s)" % test_value))))

    def test_check_uuid4(self):
        code = """
            fake_uuid = uuid.uuid4()
               """
        errors = [(1, 0, 'M354')]
        self._assert_has_errors(code, checks.check_uuid4,
                                expected_errors=errors)

        code = """
               hex_uuid = uuid.uuid4().hex
               """
        self._assert_has_no_errors(code, checks.check_uuid4)

    @ddt.unpack
    @ddt.data(
        (1, 'import mock'),
        (0, 'from unittest import mock'),
        (1, 'from mock import patch'),
        (0, 'from unittest.mock import patch'))
    def test_no_third_party_mock(self, err_count, line):
        self.assertEqual(err_count,
                         len(list(checks.no_third_party_mock(line))))

    def test_no_log_warn_check(self):
        self.assertEqual(0, len(list(checks.no_log_warn_check(
            "LOG.warning('This should not trigger LOG.warn "
            "hacking check.')"))))
        self.assertEqual(1, len(list(checks.no_log_warn_check(
            "LOG.warn('We should not use LOG.warn')"))))
        foo = """
            LOG.warn('Catch me too, please'
            )
            """
        self.assertEqual(1, len(list(checks.no_log_warn_check(foo))))



# Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Unit tests for Oracle's ZFSSA REST API.
"""
from unittest import mock

from manila import exception
from manila.share.drivers.zfssa import restclient
from manila.share.drivers.zfssa import zfssarest
from manila import test
from manila.tests import fake_zfssa


class ZFSSAApiTestCase(test.TestCase):
    """Tests ZFSSAApi."""

    @mock.patch.object(zfssarest, 'factory_restclient')
    def setUp(self, _restclient):
        super(ZFSSAApiTestCase, self).setUp()
        self.host = 'fakehost'
        self.user = 'fakeuser'
        self.url = None
        self.pool = 'fakepool'
        self.project = 'fakeproject'
        self.share = 'fakeshare'
        self.snap = 'fakesnapshot'
        _restclient.return_value = fake_zfssa.FakeRestClient()
        self._zfssa = zfssarest.ZFSSAApi()
        self._zfssa.set_host('fakehost')

        self.schema = {
            'property': 'manila_managed',
            'description': 'Managed by Manila',
            'type': 'Boolean',
        }

    def _create_response(self, status):
        response = fake_zfssa.FakeResponse(status)
        return response

    def test_enable_service(self):
        self.mock_object(self._zfssa.rclient, 'put')
        self._zfssa.rclient.put.return_value = self._create_response(
            restclient.Status.ACCEPTED)

        self._zfssa.enable_service('nfs')
        self.assertEqual(1, self._zfssa.rclient.put.call_count)

        self._zfssa.rclient.put.return_value = self._create_response(
            restclient.Status.OK)
        self.assertRaises(exception.ShareBackendException,
                          self._zfssa.enable_service,
                          'nfs')

    def test_verify_avail_space(self):
        self.mock_object(self._zfssa, 'verify_project')
        self.mock_object(self._zfssa, 'get_project_stats')
        self._zfssa.get_project_stats.return_value = 2000

        self._zfssa.verify_avail_space(self.pool,
                                       self.project,
                                       self.share,
                                       1000)
        self.assertEqual(1, self._zfssa.verify_project.call_count)
        self.assertEqual(1, self._zfssa.get_project_stats.call_count)
        self._zfssa.verify_project.assert_called_with(self.pool, self.project)
        self._zfssa.get_project_stats.assert_called_with(self.pool,
                                                         self.project)

        self._zfssa.get_project_stats.return_value = 900
        self.assertRaises(exception.ShareBackendException,
                          self._zfssa.verify_avail_space,
                          self.pool,
                          self.project,
                          self.share,
                          1000)

    def test_create_project(self):
        self.mock_object(self._zfssa, 'verify_pool')
        self.mock_object(self._zfssa.rclient, 'get')
        self.mock_object(self._zfssa.rclient, 'post')
        arg = {
            'name': self.project,
            'sharesmb': 'off',
            'sharenfs': 'off',
            'mountpoint': 'fakemnpt',
        }
        self._zfssa.rclient.get.return_value = self._create_response(
            restclient.Status.NOT_FOUND)
        self._zfssa.rclient.post.return_value = self._create_response(
            restclient.Status.CREATED)

        self._zfssa.create_project(self.pool, self.project, arg)
        self.assertEqual(1, self._zfssa.rclient.get.call_count)
        self.assertEqual(1, self._zfssa.rclient.post.call_count)
        self.assertEqual(1, self._zfssa.verify_pool.call_count)
        self._zfssa.verify_pool.assert_called_with(self.pool)

        self._zfssa.rclient.post.return_value = self._create_response(
            restclient.Status.NOT_FOUND)
        self.assertRaises(exception.ShareBackendException,
                          self._zfssa.create_project,
                          self.pool,
                          self.project,
                          arg)

    def test_create_share(self):
        self.mock_object(self._zfssa, 'verify_avail_space')
        self.mock_object(self._zfssa.rclient, 'get')
        self.mock_object(self._zfssa.rclient, 'post')
        self._zfssa.rclient.get.return_value = self._create_response(
            restclient.Status.NOT_FOUND)
        self._zfssa.rclient.post.return_value = self._create_response(
            restclient.Status.CREATED)
        arg = {
            "name": self.share,
            "quota": 1,
        }

        self._zfssa.create_share(self.pool, self.project, arg)
        self.assertEqual(1, self._zfssa.rclient.get.call_count)
        self.assertEqual(1, self._zfssa.rclient.post.call_count)
        self.assertEqual(1, self._zfssa.verify_avail_space.call_count)
        self._zfssa.verify_avail_space.assert_called_with(self.pool,
                                                          self.project,
                                                          arg,
                                                          arg['quota'])

        self._zfssa.rclient.post.return_value = self._create_response(
            restclient.Status.NOT_FOUND)
        self.assertRaises(exception.ShareBackendException,
                          self._zfssa.create_share,
                          self.pool,
                          self.project,
                          arg)

        self._zfssa.rclient.get.return_value = self._create_response(
            restclient.Status.OK)
        self.assertRaises(exception.ShareBackendException,
                          self._zfssa.create_share,
                          self.pool,
                          self.project,
                          arg)

    def test_modify_share(self):
        self.mock_object(self._zfssa.rclient, 'put')
        self._zfssa.rclient.put.return_value = self._create_response(
            restclient.Status.ACCEPTED)
        arg = {"name": "dummyname"}
        svc = self._zfssa.share_path % (self.pool, self.project, self.share)

        self._zfssa.modify_share(self.pool, self.project, self.share, arg)
        self.assertEqual(1, self._zfssa.rclient.put.call_count)
        self._zfssa.rclient.put.assert_called_with(svc, arg)

        self._zfssa.rclient.put.return_value = self._create_response(
            restclient.Status.BAD_REQUEST)
        self.assertRaises(exception.ShareBackendException,
                          self._zfssa.modify_share,
                          self.pool,
                          self.project,
                          self.share,
                          arg)

    def test_delete_share(self):
        self.mock_object(self._zfssa.rclient, 'delete')
        self._zfssa.rclient.delete.return_value = self._create_response(
            restclient.Status.NO_CONTENT)
        svc = self._zfssa.share_path % (self.pool, self.project, self.share)

        self._zfssa.delete_share(self.pool, self.project, self.share)
        self.assertEqual(1, self._zfssa.rclient.delete.call_count)
        self._zfssa.rclient.delete.assert_called_with(svc)

    def test_create_snapshot(self):
        self.mock_object(self._zfssa.rclient, 'post')
        self._zfssa.rclient.post.return_value = self._create_response(
            restclient.Status.CREATED)
        arg = {"name": self.snap}
        svc = self._zfssa.snapshots_path % (self.pool,
                                            self.project,
                                            self.share)

        self._zfssa.create_snapshot(self.pool,
                                    self.project,
                                    self.share,
                                    self.snap)
        self.assertEqual(1, self._zfssa.rclient.post.call_count)
        self._zfssa.rclient.post.assert_called_with(svc, arg)

        self._zfssa.rclient.post.return_value = self._create_response(
            restclient.Status.BAD_REQUEST)
        self.assertRaises(exception.ShareBackendException,
                          self._zfssa.create_snapshot,
                          self.pool,
                          self.project,
                          self.share,
                          self.snap)

    def test_delete_snapshot(self):
        self.mock_object(self._zfssa.rclient, 'delete')
        self._zfssa.rclient.delete.return_value = self._create_response(
            restclient.Status.NO_CONTENT)
        svc = self._zfssa.snapshot_path % (self.pool,
                                           self.project,
                                           self.share,
                                           self.snap)

        self._zfssa.delete_snapshot(self.pool,
                                    self.project,
                                    self.share,
                                    self.snap)
        self.assertEqual(1, self._zfssa.rclient.delete.call_count)
        self._zfssa.rclient.delete.assert_called_with(svc)

        self._zfssa.rclient.delete.return_value = self._create_response(
            restclient.Status.BAD_REQUEST)
        self.assertRaises(exception.ShareBackendException,
                          self._zfssa.delete_snapshot,
                          self.pool,
                          self.project,
                          self.share,
                          self.snap)

    def test_clone_snapshot(self):
        self.mock_object(self._zfssa, 'verify_avail_space')
        self.mock_object(self._zfssa.rclient, 'put')
        self._zfssa.rclient.put.return_value = self._create_response(
            restclient.Status.CREATED)
        snapshot = {
            "id": self.snap,
            "share_id": self.share,
        }
        clone = {
            "id": "cloneid",
            "size": 1,
        }
        arg = {
            "name": "dummyname",
            "quota": 1,
        }

        self._zfssa.clone_snapshot(self.pool,
                                   self.project,
                                   snapshot,
                                   clone,
                                   arg)
        self.assertEqual(1, self._zfssa.rclient.put.call_count)
        self.assertEqual(1, self._zfssa.verify_avail_space.call_count)
        self._zfssa.verify_avail_space.assert_called_with(self.pool,
                                                          self.project,
                                                          clone['id'],
                                                          clone['size'])

        self._zfssa.rclient.put.return_value = self._create_response(
            restclient.Status.NOT_FOUND)
        self.assertRaises(exception.ShareBackendException,
                          self._zfssa.clone_snapshot,
                          self.pool,
                          self.project,
                          snapshot,
                          clone,
                          arg)

    def _create_entry(self, sharenfs, ip):
        if sharenfs == 'off':
            sharenfs = 'sec=sys'
        entry = (',rw=@%s' % ip)
        if '/' not in ip:
            entry = entry + '/32'
        arg = {'sharenfs': sharenfs + entry}
        return arg

    def test_allow_access_nfs(self):
        self.mock_object(self._zfssa, 'get_share')
        self.mock_object(self._zfssa, 'modify_share')
        details = {"sharenfs": "off"}
        access = {
            "access_type": "nonip",
            "access_to": "foo",
        }

        # invalid access type
        self.assertRaises(exception.InvalidShareAccess,
                          self._zfssa.allow_access_nfs,
                          self.pool,
                          self.project,
                          self.share,
                          access)

        # valid entry
        access.update({"access_type": "ip"})
        arg = self._create_entry("off", access['access_to'])
        self._zfssa.get_share.return_value = details
        self._zfssa.allow_access_nfs(self.pool,
                                     self.project,
                                     self.share,
                                     access)
        self.assertEqual(1, self._zfssa.get_share.call_count)
        self.assertEqual(1, self._zfssa.modify_share.call_count)
        self._zfssa.get_share.assert_called_with(self.pool,
                                                 self.project,
                                                 self.share)
        self._zfssa.modify_share.assert_called_with(self.pool,
                                                    self.project,
                                                    self.share,
                                                    arg)

        # add another entry
        access.update({"access_to": "10.0.0.1/24"})
        arg = self._create_entry("off", access['access_to'])
        self._zfssa.allow_access_nfs(self.pool,
                                     self.project,
                                     self.share,
                                     access)
        self.assertEqual(2, self._zfssa.modify_share.call_count)
        self._zfssa.modify_share.assert_called_with(self.pool,
                                                    self.project,
                                                    self.share,
                                                    arg)

        # verify modify_share is not called if sharenfs='on'
        details = {"sharenfs": "on"}
        self._zfssa.get_share.return_value = details
        self._zfssa.allow_access_nfs(self.pool,
                                     self.project,
                                     self.share,
                                     access)
        self.assertEqual(2, self._zfssa.modify_share.call_count)

        # verify modify_share is not called if ip is already in the list
        access.update({"access_to": "10.0.0.1/24"})
        details = self._create_entry("off", access['access_to'])
        self._zfssa.get_share.return_value = details
        self._zfssa.allow_access_nfs(self.pool,
                                     self.project,
                                     self.share,
                                     access)
        self.assertEqual(2, self._zfssa.modify_share.call_count)

    def test_deny_access_nfs(self):
        self.mock_object(self._zfssa, 'get_share')
        self.mock_object(self._zfssa, 'modify_share')
        data1 = self._create_entry("off", "10.0.0.1")
        access = {
            "access_type": "nonip",
            "access_to": "foo",
        }

        # invalid access_type
        self.assertRaises(exception.InvalidShareAccess,
                          self._zfssa.deny_access_nfs,
                          self.pool,
                          self.project,
                          self.share,
                          access)

        # valid entry
        access.update({"access_type": "ip"})
        self._zfssa.get_share.return_value = data1
        self._zfssa.deny_access_nfs(self.pool,
                                    self.project,
                                    self.share,
                                    access)
        self.assertEqual(1, self._zfssa.get_share.call_count)
        self.assertEqual(0, self._zfssa.modify_share.call_count)
        self._zfssa.get_share.assert_called_with(self.pool,
                                                 self.project,
                                                 self.share)
        # another valid entry
        data1 = self._create_entry(data1['sharenfs'], '10.0.0.2/24')
        data2 = self._create_entry(data1['sharenfs'], access['access_to'])
        self._zfssa.get_share.return_value = data2
        self._zfssa.deny_access_nfs(self.pool,
                                    self.project,
                                    self.share,
                                    access)
        self.assertEqual(2, self._zfssa.get_share.call_count)
        self.assertEqual(1, self._zfssa.modify_share.call_count)
        self._zfssa.get_share.assert_called_with(self.pool,
                                                 self.project,
                                                 self.share)
        self._zfssa.modify_share.assert_called_with(self.pool,
                                                    self.project,
                                                    self.share,
                                                    data1)

    def test_create_schema_negative(self):
        self.mock_object(self._zfssa.rclient, 'get')
        self.mock_object(self._zfssa.rclient, 'post')
        self._zfssa.rclient.post.return_value = self._create_response(
            restclient.Status.NOT_FOUND)

        self.assertRaises(exception.ShareBackendException,
                          self._zfssa.create_schema,
                          self.schema)

    def test_create_schema_property_exists(self):
        self.mock_object(self._zfssa.rclient, 'get')
        self.mock_object(self._zfssa.rclient, 'post')
        self._zfssa.rclient.get.return_value = self._create_response(
            restclient.Status.OK)

        self._zfssa.create_schema(self.schema)

        self.assertEqual(1, self._zfssa.rclient.get.call_count)
        self.assertEqual(0, self._zfssa.rclient.post.call_count)

    def test_create_schema(self):
        self.mock_object(self._zfssa.rclient, 'get')
        self.mock_object(self._zfssa.rclient, 'post')
        self._zfssa.rclient.get.return_value = self._create_response(
            restclient.Status.NOT_FOUND)
        self._zfssa.rclient.post.return_value = self._create_response(
            restclient.Status.CREATED)

        self._zfssa.create_schema(self.schema)

        self.assertEqual(1, self._zfssa.rclient.get.call_count)
        self.assertEqual(1, self._zfssa.rclient.post.call_count)



# Copyright 2015 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from webob import exc

from manila.api import common
from manila.api.openstack import wsgi
from manila.api.views import share_instance as instance_view
from manila import db
from manila import exception
from manila import share


class ShareInstancesController(wsgi.Controller, wsgi.AdminActionsMixin):
    """The share instances API controller for the OpenStack API."""

    resource_name = 'share_instance'
    _view_builder_class = instance_view.ViewBuilder

    def __init__(self):
        self.share_api = share.API()
        super(ShareInstancesController, self).__init__()

    def _get(self, *args, **kwargs):
        return db.share_instance_get(*args, **kwargs)

    def _update(self, *args, **kwargs):
        db.share_instance_update(*args, **kwargs)

    def _delete(self, *args, **kwargs):
        return self.share_api.delete_instance(*args, **kwargs)

    @wsgi.Controller.api_version('2.3', '2.6')
    @wsgi.action('os-reset_status')
    def instance_reset_status_legacy(self, req, id, body):
        return self._reset_status(req, id, body)

    @wsgi.Controller.api_version('2.7')
    @wsgi.action('reset_status')
    def instance_reset_status(self, req, id, body):
        return self._reset_status(req, id, body)

    @wsgi.Controller.api_version('2.3', '2.6')
    @wsgi.action('os-force_delete')
    def instance_force_delete_legacy(self, req, id, body):
        return self._force_delete(req, id, body)

    @wsgi.Controller.api_version('2.7')
    @wsgi.action('force_delete')
    def instance_force_delete(self, req, id, body):
        return self._force_delete(req, id, body)

    @wsgi.Controller.api_version("2.3", "2.34")  # noqa
    @wsgi.Controller.authorize
    def index(self, req):  # pylint: disable=function-redefined
        context = req.environ['manila.context']

        req.GET.pop('export_location_id', None)
        req.GET.pop('export_location_path', None)
        instances = db.share_instances_get_all(context)
        return self._view_builder.detail_list(req, instances)

    @wsgi.Controller.api_version("2.35")  # noqa
    @wsgi.Controller.authorize
    def index(self, req):  # pylint: disable=function-redefined  # noqa F811
        context = req.environ['manila.context']
        filters = {}
        filters.update(req.GET)
        common.remove_invalid_options(
            context, filters, ('export_location_id', 'export_location_path'))

        instances = db.share_instances_get_all(context, filters)
        return self._view_builder.detail_list(req, instances)

    @wsgi.Controller.api_version("2.3")
    @wsgi.Controller.authorize
    def show(self, req, id):
        context = req.environ['manila.context']

        try:
            instance = db.share_instance_get(context, id)
        except exception.NotFound:
            raise exc.HTTPNotFound()

        return self._view_builder.detail(req, instance)

    @wsgi.Controller.api_version("2.3")
    @wsgi.Controller.authorize('index')
    def get_share_instances(self, req, share_id):
        context = req.environ['manila.context']

        try:
            share = self.share_api.get(context, share_id)
        except exception.NotFound:
            raise exc.HTTPNotFound()

        view = instance_view.ViewBuilder()
        return view.detail_list(req, share.instances)


def create_resource():
    return wsgi.Resource(ShareInstancesController())



# Copyright 2011 OpenStack LLC.
# Copyright 2011 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright (c) 2015 Mirantis inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
WSGI middleware for OpenStack Share API v2.
"""

from manila.api import extensions
import manila.api.openstack
from manila.api.v1 import limits
from manila.api.v1 import scheduler_stats
from manila.api.v1 import security_service
from manila.api.v1 import share_manage
from manila.api.v1 import share_metadata
from manila.api.v1 import share_types_extra_specs
from manila.api.v1 import share_unmanage
from manila.api.v2 import availability_zones
from manila.api.v2 import messages
from manila.api.v2 import quota_class_sets
from manila.api.v2 import quota_sets
from manila.api.v2 import services
from manila.api.v2 import share_access_metadata
from manila.api.v2 import share_accesses
from manila.api.v2 import share_export_locations
from manila.api.v2 import share_group_snapshots
from manila.api.v2 import share_group_type_specs
from manila.api.v2 import share_group_types
from manila.api.v2 import share_groups
from manila.api.v2 import share_instance_export_locations
from manila.api.v2 import share_instances
from manila.api.v2 import share_network_subnets
from manila.api.v2 import share_networks
from manila.api.v2 import share_replica_export_locations
from manila.api.v2 import share_replicas
from manila.api.v2 import share_servers
from manila.api.v2 import share_snapshot_export_locations
from manila.api.v2 import share_snapshot_instance_export_locations
from manila.api.v2 import share_snapshot_instances
from manila.api.v2 import share_snapshots
from manila.api.v2 import share_types
from manila.api.v2 import shares
from manila.api import versions


class APIRouter(manila.api.openstack.APIRouter):
    """Route API requests.

    Routes requests on the OpenStack API to the appropriate controller
    and method.
    """
    ExtensionManager = extensions.ExtensionManager

    def _setup_routes(self, mapper):
        self.resources["versions"] = versions.create_resource()
        mapper.connect("versions", "/",
                       controller=self.resources["versions"],
                       action="index")

        mapper.redirect("", "/")

        self.resources["availability_zones_legacy"] = (
            availability_zones.create_resource_legacy())
        # TODO(vponomaryov): "os-availability-zone" is deprecated
        # since v2.7. Remove it when minimum API version becomes equal to
        # or greater than v2.7.
        mapper.resource("availability-zone",
                        "os-availability-zone",
                        controller=self.resources["availability_zones_legacy"])

        self.resources["availability_zones"] = (
            availability_zones.create_resource())
        mapper.resource("availability-zone",
                        "availability-zones",
                        controller=self.resources["availability_zones"])

        self.resources["services_legacy"] = services.create_resource_legacy()
        # TODO(vponomaryov): "os-services" is deprecated
        # since v2.7. Remove it when minimum API version becomes equal to
        # or greater than v2.7.
        mapper.resource("service",
                        "os-services",
                        controller=self.resources["services_legacy"])

        self.resources["services"] = services.create_resource()
        mapper.resource("service",
                        "services",
                        controller=self.resources["services"])

        self.resources["quota_sets_legacy"] = (
            quota_sets.create_resource_legacy())
        # TODO(vponomaryov): "os-quota-sets" is deprecated
        # since v2.7. Remove it when minimum API version becomes equal to
        # or greater than v2.7.
        mapper.resource("quota-set",
                        "os-quota-sets",
                        controller=self.resources["quota_sets_legacy"],
                        member={"defaults": "GET"})

        self.resources["quota_sets"] = quota_sets.create_resource()
        mapper.resource("quota-set",
                        "quota-sets",
                        controller=self.resources["quota_sets"],
                        member={"defaults": "GET",
                                "detail": "GET"})

        self.resources["quota_class_sets_legacy"] = (
            quota_class_sets.create_resource_legacy())
        # TODO(vponomaryov): "os-quota-class-sets" is deprecated
        # since v2.7. Remove it when minimum API version becomes equal to
        # or greater than v2.7.
        mapper.resource("quota-class-set",
                        "os-quota-class-sets",
                        controller=self.resources["quota_class_sets_legacy"])

        self.resources["quota_class_sets"] = quota_class_sets.create_resource()
        mapper.resource("quota-class-set",
                        "quota-class-sets",
                        controller=self.resources["quota_class_sets"])

        self.resources["share_manage"] = share_manage.create_resource()
        # TODO(vponomaryov): "os-share-manage" is deprecated
        # since v2.7. Remove it when minimum API version becomes equal to
        # or greater than v2.7.
        mapper.resource("share_manage",
                        "os-share-manage",
                        controller=self.resources["share_manage"])

        self.resources["share_unmanage"] = share_unmanage.create_resource()
        # TODO(vponomaryov): "os-share-unmanage" is deprecated
        # since v2.7. Remove it when minimum API version becomes equal to
        # or greater than v2.7.
        mapper.resource("share_unmanage",
                        "os-share-unmanage",
                        controller=self.resources["share_unmanage"],
                        member={"unmanage": "POST"})

        self.resources["shares"] = shares.create_resource()
        mapper.resource("share", "shares",
                        controller=self.resources["shares"],
                        collection={"detail": "GET"},
                        member={"action": "POST"})

        for path_prefix in ['/{project_id}', '']:
            # project_id is optional
            mapper.connect("shares",
                           "%s/shares/manage" % path_prefix,
                           controller=self.resources["shares"],
                           action="manage",
                           conditions={"method": ["POST"]})

        self.resources["share_instances"] = share_instances.create_resource()
        mapper.resource("share_instance", "share_instances",
                        controller=self.resources["share_instances"],
                        collection={"detail": "GET"},
                        member={"action": "POST"})

        self.resources["share_instance_export_locations"] = (
            share_instance_export_locations.create_resource())

        for path_prefix in ['/{project_id}', '']:
            # project_id is optional
            mapper.connect("share_instances",
                           ("%s/share_instances/{share_instance_id}"
                            "/export_locations" % path_prefix),
                           controller=self.resources[
                               "share_instance_export_locations"],
                           action="index",
                           conditions={"method": ["GET"]})

            mapper.connect("share_instances",
                           ("%s/share_instances/{share_instance_id}"
                            "/export_locations"
                            "/{export_location_uuid}" % path_prefix),
                           controller=self.resources[
                               "share_instance_export_locations"],
                           action="show",
                           conditions={"method": ["GET"]})

            mapper.connect("share_instance",
                           "%s/shares/{share_id}/instances" % path_prefix,
                           controller=self.resources["share_instances"],
                           action="get_share_instances",
                           conditions={"method": ["GET"]})

            self.resources["share_export_locations"] = (
                share_export_locations.create_resource())
            mapper.connect("shares",
                           "%s/shares/{share_id}"
                           "/export_locations" % path_prefix,
                           controller=self.resources["share_export_locations"],
                           action="index",
                           conditions={"method": ["GET"]})
            mapper.connect("shares",
                           ("%s/shares/{share_id}/export_locations"
                            "/{export_location_uuid}" % path_prefix),
                           controller=self.resources["share_export_locations"],
                           action="show",
                           conditions={"method": ["GET"]})

        self.resources["snapshots"] = share_snapshots.create_resource()
        mapper.resource("snapshot", "snapshots",
                        controller=self.resources["snapshots"],
                        collection={"detail": "GET"},
                        member={"action": "POST"})

        for path_prefix in ['/{project_id}', '']:
            # project_id is optional
            mapper.connect("snapshots",
                           "%s/snapshots/manage" % path_prefix,
                           controller=self.resources["snapshots"],
                           action="manage",
                           conditions={"method": ["POST"]})

            mapper.connect("snapshots",
                           "%s/snapshots/{snapshot_id}"
                           "/access-list" % path_prefix,
                           controller=self.resources["snapshots"],
                           action="access_list",
                           conditions={"method": ["GET"]})

            self.resources["share_snapshot_export_locations"] = (
                share_snapshot_export_locations.create_resource())
            mapper.connect("snapshots",
                           "%s/snapshots/{snapshot_id}"
                           "/export-locations" % path_prefix,
                           controller=self.resources[
                               "share_snapshot_export_locations"],
                           action="index",
                           conditions={"method": ["GET"]})

            mapper.connect("snapshots",
                           "%s/snapshots/{snapshot_id}/export-locations"
                           "/{export_location_id}" % path_prefix,
                           controller=self.resources[
                               "share_snapshot_export_locations"],
                           action="show",
                           conditions={"method": ["GET"]})

        self.resources['snapshot_instances'] = (
            share_snapshot_instances.create_resource())
        mapper.resource("snapshot-instance", "snapshot-instances",
                        controller=self.resources['snapshot_instances'],
                        collection={'detail': 'GET'},
                        member={'action': 'POST'})

        self.resources["share_snapshot_instance_export_locations"] = (
            share_snapshot_instance_export_locations.create_resource())

        for path_prefix in ['/{project_id}', '']:
            # project_id is optional
            mapper.connect("snapshot-instance",
                           "%s/snapshot-instances/{snapshot_instance_id}"
                           "/export-locations" % path_prefix,
                           controller=self.resources[
                               "share_snapshot_instance_export_locations"],
                           action="index",
                           conditions={"method": ["GET"]})

            mapper.connect("snapshot-instance",
                           "%s/snapshot-instances/{snapshot_instance_id}"
                           "/export-locations"
                           "/{export_location_id}" % path_prefix,
                           controller=self.resources[
                               "share_snapshot_instance_export_locations"],
                           action="show",
                           conditions={"method": ["GET"]})

        self.resources["share_metadata"] = share_metadata.create_resource()
        share_metadata_controller = self.resources["share_metadata"]

        mapper.resource("share_metadata", "metadata",
                        controller=share_metadata_controller,
                        parent_resource=dict(member_name="share",
                                             collection_name="shares"))

        for path_prefix in ['/{project_id}', '']:
            # project_id is optional
            mapper.connect("metadata",
                           "%s/shares/{share_id}/metadata" % path_prefix,
                           controller=share_metadata_controller,
                           action="update_all",
                           conditions={"method": ["PUT"]})

        self.resources["limits"] = limits.create_resource()
        mapper.resource("limit", "limits",
                        controller=self.resources["limits"])

        self.resources["security_services"] = (
            security_service.create_resource())
        mapper.resource("security-service", "security-services",
                        controller=self.resources["security_services"],
                        collection={"detail": "GET"})

        self.resources["share_networks"] = share_networks.create_resource()
        mapper.resource(share_networks.RESOURCE_NAME,
                        "share-networks",
                        controller=self.resources["share_networks"],
                        collection={"detail": "GET"},
                        member={"action": "POST"})

        for path_prefix in ['/{project_id}', '']:
            # project_id is optional
            self.resources["share_network_subnets"] = (
                share_network_subnets.create_resource())
            mapper.connect("share-networks",
                           "%s/share-networks/{share_network_id}"
                           "/subnets" % path_prefix,
                           controller=self.resources["share_network_subnets"],
                           action="create",
                           conditions={"method": ["POST"]})
            mapper.connect("share-networks",
                           "%s/share-networks/{share_network_id}"
                           "/subnets/{share_network_subnet_id}" % path_prefix,
                           controller=self.resources["share_network_subnets"],
                           action="delete",
                           conditions={"method": ["DELETE"]})
            mapper.connect("share-networks",
                           "%s/share-networks/{share_network_id}"
                           "/subnets/{share_network_subnet_id}" % path_prefix,
                           controller=self.resources["share_network_subnets"],
                           action="show",
                           conditions={"method": ["GET"]})
            mapper.connect("share-networks",
                           "%s/share-networks/{share_network_id}"
                           "/subnets" % path_prefix,
                           controller=self.resources["share_network_subnets"],
                           action="index",
                           conditions={"method": ["GET"]})

        self.resources["share_servers"] = share_servers.create_resource()
        mapper.resource("share_server",
                        "share-servers",
                        controller=self.resources["share_servers"],
                        member={"action": "POST"})

        for path_prefix in ['/{project_id}', '']:
            # project_id is optional
            mapper.connect("details",
                           "%s/share-servers/{id}/details" % path_prefix,
                           controller=self.resources["share_servers"],
                           action="details",
                           conditions={"method": ["GET"]})
            mapper.connect("share_servers",
                           "%s/share-servers/manage" % path_prefix,
                           controller=self.resources["share_servers"],
                           action="manage",
                           conditions={"method": ["POST"]})

        self.resources["types"] = share_types.create_resource()
        mapper.resource("type", "types",
                        controller=self.resources["types"],
                        collection={"detail": "GET", "default": "GET"},
                        member={"action": "POST",
                                "os-share-type-access": "GET",
                                "share_type_access": "GET"})

        self.resources["extra_specs"] = (
            share_types_extra_specs.create_resource())
        mapper.resource("extra_spec", "extra_specs",
                        controller=self.resources["extra_specs"],
                        parent_resource=dict(member_name="type",
                                             collection_name="types"))

        self.resources["scheduler_stats"] = scheduler_stats.create_resource()
        for path_prefix in ['/{project_id}', '']:
            # project_id is optional
            mapper.connect("pools",
                           "%s/scheduler-stats/pools" % path_prefix,
                           controller=self.resources["scheduler_stats"],
                           action="pools_index",
                           conditions={"method": ["GET"]})
            mapper.connect("pools",
                           "%s/scheduler-stats/pools/detail" % path_prefix,
                           controller=self.resources["scheduler_stats"],
                           action="pools_detail",
                           conditions={"method": ["GET"]})

        self.resources["share-groups"] = share_groups.create_resource()
        mapper.resource(
            "share-group",
            "share-groups",
            controller=self.resources["share-groups"],
            collection={"detail": "GET"})
        for path_prefix in ['/{project_id}', '']:
            # project_id is optional
            mapper.connect(
                "share-groups",
                "%s/share-groups/{id}/action" % path_prefix,
                controller=self.resources["share-groups"],
                action="action",
                conditions={"method": ["POST"]})

        self.resources["share-group-types"] = (
            share_group_types.create_resource())
        mapper.resource(
            "share-group-type",
            "share-group-types",
            controller=self.resources["share-group-types"],
            collection={"detail": "GET", "default": "GET"},
            member={"action": "POST"})
        for path_prefix in ['/{project_id}', '']:
            # project_id is optional
            mapper.connect(
                "share-group-types",
                "%s/share-group-types/{id}/access" % path_prefix,
                controller=self.resources["share-group-types"],
                action="share_group_type_access",
                conditions={"method": ["GET"]})

        # NOTE(ameade): These routes can be simplified when the following
        # issue is fixed: https://github.com/bbangert/routes/issues/68
        self.resources["group-specs"] = (
            share_group_type_specs.create_resource())

        for path_prefix in ['/{project_id}', '']:
            # project_id is optional
            mapper.connect(
                "share-group-types",
                "%s/share-group-types/{id}/group-specs" % path_prefix,
                controller=self.resources["group-specs"],
                action="index",
                conditions={"method": ["GET"]})
            mapper.connect(
                "share-group-types",
                "%s/share-group-types/{id}/group-specs" % path_prefix,
                controller=self.resources["group-specs"],
                action="create",
                conditions={"method": ["POST"]})
            mapper.connect(
                "share-group-types",
                "%s/share-group-types/{id}/group-specs/{key}" % path_prefix,
                controller=self.resources["group-specs"],
                action="show",
                conditions={"method": ["GET"]})
            mapper.connect(
                "share-group-types",
                "%s/share-group-types/{id}/group-specs/{key}" % path_prefix,
                controller=self.resources["group-specs"],
                action="delete",
                conditions={"method": ["DELETE"]})
            mapper.connect(
                "share-group-types",
                "%s/share-group-types/{id}/group-specs/{key}" % path_prefix,
                controller=self.resources["group-specs"],
                action="update",
                conditions={"method": ["PUT"]})

        self.resources["share-group-snapshots"] = (
            share_group_snapshots.create_resource())
        mapper.resource(
            "share-group-snapshot",
            "share-group-snapshots",
            controller=self.resources["share-group-snapshots"],
            collection={"detail": "GET"},
            member={"members": "GET", "action": "POST"})

        for path_prefix in ['/{project_id}', '']:
            # project_id is optional
            mapper.connect(
                "share-group-snapshots",
                "%s/share-group-snapshots/{id}/action" % path_prefix,
                controller=self.resources["share-group-snapshots"],
                action="action",
                conditions={"method": ["POST"]})

        self.resources['share-replicas'] = share_replicas.create_resource()
        mapper.resource("share-replica", "share-replicas",
                        controller=self.resources['share-replicas'],
                        collection={'detail': 'GET'},
                        member={'action': 'POST'})

        self.resources["share-replica-export-locations"] = (
            share_replica_export_locations.create_resource())
        for path_prefix in ['/{project_id}', '']:
            # project_id is optional
            mapper.connect("share-replicas",
                           ("%s/share-replicas/{share_replica_id}"
                            "/export-locations" % path_prefix),
                           controller=self.resources[
                               "share-replica-export-locations"],
                           action="index",
                           conditions={"method": ["GET"]})
            mapper.connect("share-replicas",
                           ("%s/share-replicas/{share_replica_id}"
                            "/export-locations"
                            "/{export_location_uuid}" % path_prefix),
                           controller=self.resources[
                               "share-replica-export-locations"],
                           action="show",
                           conditions={"method": ["GET"]})

        self.resources['messages'] = messages.create_resource()
        mapper.resource("message", "messages",
                        controller=self.resources['messages'])

        self.resources["share-access-rules"] = share_accesses.create_resource()
        mapper.resource(
            "share-access-rule",
            "share-access-rules",
            controller=self.resources["share-access-rules"],
            collection={"detail": "GET"})

        for path_prefix in ['/{project_id}', '']:
            # project_id is optional
            self.resources["access-metadata"] = (
                share_access_metadata.create_resource())
            access_metadata_controller = self.resources["access-metadata"]
            mapper.connect("share-access-rules",
                           "%s/share-access-rules"
                           "/{access_id}/metadata" % path_prefix,
                           controller=access_metadata_controller,
                           action="update",
                           conditions={"method": ["PUT"]})

            mapper.connect("share-access-rules",
                           "%s/share-access-rules"
                           "/{access_id}/metadata/{key}" % path_prefix,
                           controller=access_metadata_controller,
                           action="delete",
                           conditions={"method": ["DELETE"]})



# Copyright (c) 2017 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add description for share type

Revision ID: 27cb96d991fa
Revises: 829a09b0ddd4
Create Date: 2017-09-16 03:07:15.548947

"""

# revision identifiers, used by Alembic.
revision = '27cb96d991fa'
down_revision = '829a09b0ddd4'

from alembic import op
from oslo_log import log
import sqlalchemy as sa

LOG = log.getLogger(__name__)


def upgrade():
    try:
        op.add_column(
            'share_types',
            sa.Column('description', sa.String(255), nullable=True))
    except Exception:
        LOG.error("Column share_types.description not created!")
        raise


def downgrade():
    try:
        op.drop_column('share_types', 'description')
    except Exception:
        LOG.error("Column share_types.description not dropped!")
        raise
